#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import itertools
import os
import unittest
import numpy

from openfermion.config import DATA_DIRECTORY
from openfermion.chem import MolecularData
from openfermion.ops.operators import FermionOperator
from openfermion.transforms.opconversions import (get_fermion_operator,
                                                  normal_ordered)
from openfermion.linalg import eigenspectrum
from openfermion.utils.operator_utils import is_hermitian
from openfermion.testing.testing_utils import random_interaction_operator
from openfermion.circuits.low_rank import (get_chemist_two_body_coefficients,
                                           low_rank_two_body_decomposition,
                                           prepare_one_body_squared_evolution)


class ChemistTwoBodyTest(unittest.TestCase):

    def test_operator_consistency_w_spin(self):

        # Initialize a random InteractionOperator and FermionOperator.
        n_orbitals = 3
        n_qubits = 2 * n_orbitals
        random_interaction = random_interaction_operator(n_orbitals,
                                                         expand_spin=True,
                                                         real=False,
                                                         seed=8)
        random_fermion = get_fermion_operator(random_interaction)

        # Convert to chemist ordered tensor.
        one_body_correction, chemist_tensor = \
            get_chemist_two_body_coefficients(
                random_interaction.two_body_tensor, spin_basis=True)

        # Convert output to FermionOperator.
        output_operator = FermionOperator((), random_interaction.constant)

        # Convert one-body.
        one_body_coefficients = (random_interaction.one_body_tensor +
                                 one_body_correction)
        for p, q in itertools.product(range(n_qubits), repeat=2):
            term = ((p, 1), (q, 0))
            coefficient = one_body_coefficients[p, q]
            output_operator += FermionOperator(term, coefficient)

        # Convert two-body.
        for p, q, r, s in itertools.product(range(n_orbitals), repeat=4):
            coefficient = chemist_tensor[p, q, r, s]

            # Mixed spin.
            term = ((2 * p, 1), (2 * q, 0), (2 * r + 1, 1), (2 * s + 1, 0))
            output_operator += FermionOperator(term, coefficient)

            term = ((2 * p + 1, 1), (2 * q + 1, 0), (2 * r, 1), (2 * s, 0))
            output_operator += FermionOperator(term, coefficient)

            # Same spin.
            term = ((2 * p, 1), (2 * q, 0), (2 * r, 1), (2 * s, 0))
            output_operator += FermionOperator(term, coefficient)

            term = ((2 * p + 1, 1), (2 * q + 1, 0), (2 * r + 1, 1), (2 * s + 1,
                                                                     0))
            output_operator += FermionOperator(term, coefficient)

        # Check that difference is small.
        difference = normal_ordered(random_fermion - output_operator)
        self.assertAlmostEqual(0., difference.induced_norm())


class LowRankTest(unittest.TestCase):

    def test_random_operator_consistency(self):

        # Initialize a random InteractionOperator and FermionOperator.
        n_orbitals = 3
        n_qubits = 2 * n_orbitals
        random_interaction = random_interaction_operator(n_orbitals,
                                                         expand_spin=True,
                                                         real=True,
                                                         seed=8)
        random_fermion = get_fermion_operator(random_interaction)

        # Decompose.
        eigenvalues, one_body_squares, one_body_correction, error = (
            low_rank_two_body_decomposition(random_interaction.two_body_tensor))
        self.assertFalse(error)

        # Build back operator constant and one-body components.
        decomposed_operator = FermionOperator((), random_interaction.constant)
        one_body_coefficients = (random_interaction.one_body_tensor +
                                 one_body_correction)
        for p, q in itertools.product(range(n_qubits), repeat=2):
            term = ((p, 1), (q, 0))
            coefficient = one_body_coefficients[p, q]
            decomposed_operator += FermionOperator(term, coefficient)

        # Build back two-body component.
        for l in range(n_orbitals**2):
            one_body_operator = FermionOperator()
            for p, q in itertools.product(range(n_qubits), repeat=2):
                term = ((p, 1), (q, 0))
                coefficient = one_body_squares[l, p, q]
                one_body_operator += FermionOperator(term, coefficient)
            decomposed_operator += eigenvalues[l] * (one_body_operator**2)

        # Test for consistency.
        difference = normal_ordered(decomposed_operator - random_fermion)
        self.assertAlmostEqual(0., difference.induced_norm())

    def test_molecular_operator_consistency(self):

        # Initialize H2 InteractionOperator.
        n_qubits = 4
        filename = os.path.join(DATA_DIRECTORY, 'H2_sto-3g_singlet_0.7414')
        molecule = MolecularData(filename=filename)
        molecule_interaction = molecule.get_molecular_hamiltonian()
        molecule_operator = get_fermion_operator(molecule_interaction)

        constant = molecule_interaction.constant
        one_body_coefficients = molecule_interaction.one_body_tensor
        two_body_coefficients = molecule_interaction.two_body_tensor

        # Perform decomposition.
        eigenvalues, one_body_squares, one_body_corrections, trunc_error = (
            low_rank_two_body_decomposition(two_body_coefficients))
        self.assertAlmostEqual(trunc_error, 0.)

        # Build back operator constant and one-body components.
        decomposed_operator = FermionOperator((), constant)
        for p, q in itertools.product(range(n_qubits), repeat=2):
            term = ((p, 1), (q, 0))
            coefficient = (one_body_coefficients[p, q] +
                           one_body_corrections[p, q])
            decomposed_operator += FermionOperator(term, coefficient)

        # Build back two-body component.
        for l in range(one_body_squares.shape[0]):
            one_body_operator = FermionOperator()
            for p, q in itertools.product(range(n_qubits), repeat=2):
                term = ((p, 1), (q, 0))
                coefficient = one_body_squares[l, p, q]
                if abs(eigenvalues[l]) > 1e-6:
                    self.assertTrue(is_hermitian(one_body_squares[l]))
                one_body_operator += FermionOperator(term, coefficient)
            decomposed_operator += eigenvalues[l] * (one_body_operator**2)

        # Test for consistency.
        difference = normal_ordered(decomposed_operator - molecule_operator)
        self.assertAlmostEqual(0., difference.induced_norm())

        # Decompose with slightly negative operator that must use eigen
        molecule = MolecularData(filename=filename)
        molecule.two_body_integrals[0, 0, 0, 0] -= 1

        eigenvalues, one_body_squares, one_body_corrections, trunc_error = (
            low_rank_two_body_decomposition(two_body_coefficients))
        self.assertAlmostEqual(trunc_error, 0.)

        # Check for property errors
        with self.assertRaises(TypeError):
            eigenvalues, one_body_squares, _, trunc_error = (
                low_rank_two_body_decomposition(two_body_coefficients + 0.01j,
                                                truncation_threshold=1.,
                                                final_rank=1))

    def test_rank_reduction(self):

        # Initialize H2 InteractionOperator.
        n_qubits = 4
        n_orbitals = 2
        filename = os.path.join(DATA_DIRECTORY, 'H2_sto-3g_singlet_0.7414')
        molecule = MolecularData(filename=filename)
        molecule_interaction = molecule.get_molecular_hamiltonian()
        fermion_operator = get_fermion_operator(molecule_interaction)

        constant = molecule_interaction.constant
        two_body_coefficients = molecule_interaction.two_body_tensor

        # Rank reduce with threshold.
        errors = []
        for truncation_threshold in [1., 0.1, 0.01, 0.001]:

            # Decompose with threshold.
            (test_eigenvalues, one_body_squares, one_body_correction,
             trunc_error) = low_rank_two_body_decomposition(
                 two_body_coefficients,
                 truncation_threshold=truncation_threshold)

            # Make sure error is below truncation specification.
            self.assertTrue(trunc_error > 0.)
            self.assertTrue(trunc_error < truncation_threshold)
            self.assertTrue(len(test_eigenvalues) < n_orbitals**2)
            self.assertTrue(len(one_body_squares) == len(test_eigenvalues))

            # Build back operator constant and one-body components.
            decomposed_operator = FermionOperator((), constant)
            one_body_coefficients = (molecule_interaction.one_body_tensor +
                                     one_body_correction)
            for p, q in itertools.product(range(n_qubits), repeat=2):
                term = ((p, 1), (q, 0))
                coefficient = one_body_coefficients[p, q]
                decomposed_operator += FermionOperator(term, coefficient)

            # Build back two-body component.
            for l in range(len(test_eigenvalues)):
                one_body_operator = FermionOperator()
                for p, q in itertools.product(range(n_qubits), repeat=2):
                    term = ((p, 1), (q, 0))
                    coefficient = one_body_squares[l, p, q]
                    one_body_operator += FermionOperator(term, coefficient)
                decomposed_operator += (test_eigenvalues[l] *
                                        one_body_operator**2)

            # Test for consistency.
            difference = normal_ordered(decomposed_operator - fermion_operator)
            errors += [difference.induced_norm()]
            self.assertTrue(errors[-1] <= trunc_error)
        self.assertTrue(errors[3] <= errors[2] <= errors[1] <= errors[0])

        # Rank reduce by imposing final rank.
        errors = []
        for final_rank in [1, 2, 3, 4]:

            # Decompose with threshold.
            (test_eigenvalues, one_body_squares, one_body_correction,
             trunc_error) = low_rank_two_body_decomposition(
                 two_body_coefficients, final_rank=final_rank)

            # Make sure error is below truncation specification.
            self.assertTrue(len(test_eigenvalues) == final_rank)

            # Build back operator constant and one-body components.
            decomposed_operator = FermionOperator((), constant)
            one_body_coefficients = (molecule_interaction.one_body_tensor +
                                     one_body_correction)
            for p, q in itertools.product(range(n_qubits), repeat=2):
                term = ((p, 1), (q, 0))
                coefficient = one_body_coefficients[p, q]
                decomposed_operator += FermionOperator(term, coefficient)

            # Build back two-body component.
            for l in range(final_rank):
                one_body_operator = FermionOperator()
                for p, q in itertools.product(range(n_qubits), repeat=2):
                    term = ((p, 1), (q, 0))
                    coefficient = one_body_squares[l, p, q]
                    one_body_operator += FermionOperator(term, coefficient)
                decomposed_operator += (test_eigenvalues[l] *
                                        one_body_operator**2)

            # Test for consistency.
            difference = normal_ordered(decomposed_operator - fermion_operator)
            errors += [difference.induced_norm()]
        self.assertTrue(errors[3] <= errors[2] <= errors[1] <= errors[0])

    def test_one_body_square_decomposition(self):

        # Initialize H2 InteractionOperator.
        n_qubits = 4
        filename = os.path.join(DATA_DIRECTORY, 'H2_sto-3g_singlet_0.7414')
        molecule = MolecularData(filename=filename)
        molecule_interaction = molecule.get_molecular_hamiltonian()
        get_fermion_operator(molecule_interaction)

        two_body_coefficients = molecule_interaction.two_body_tensor

        # Decompose.
        eigenvalues, one_body_squares, _, _ = (low_rank_two_body_decomposition(
            two_body_coefficients, truncation_threshold=0))
        rank = eigenvalues.size
        for l in range(rank):
            one_body_operator = FermionOperator()
            for p, q in itertools.product(range(n_qubits), repeat=2):
                term = ((p, 1), (q, 0))
                coefficient = one_body_squares[l, p, q]
                one_body_operator += FermionOperator(term, coefficient)
            one_body_squared = one_body_operator**2

            # Get the squared one-body operator via one-body decomposition.
            if abs(eigenvalues[l]) < 1e-6:
                with self.assertRaises(ValueError):
                    prepare_one_body_squared_evolution(one_body_squares[l])
                continue
            else:
                density_density_matrix, basis_transformation_matrix = (
                    prepare_one_body_squared_evolution(one_body_squares[l]))
            two_body_operator = FermionOperator()
            for p, q in itertools.product(range(n_qubits), repeat=2):
                term = ((p, 1), (p, 0), (q, 1), (q, 0))
                coefficient = density_density_matrix[p, q]
                two_body_operator += FermionOperator(term, coefficient)

            # Confirm that the rotations diagonalize the one-body squares.
            hopefully_diagonal = basis_transformation_matrix.dot(
                numpy.dot(
                    one_body_squares[l],
                    numpy.transpose(
                        numpy.conjugate(basis_transformation_matrix))))
            diagonal = numpy.diag(hopefully_diagonal)
            difference = hopefully_diagonal - numpy.diag(diagonal)
            self.assertAlmostEqual(0., numpy.amax(numpy.absolute(difference)))
            density_density_alternative = numpy.outer(diagonal, diagonal)
            difference = density_density_alternative - density_density_matrix
            self.assertAlmostEqual(0., numpy.amax(numpy.absolute(difference)))

            # Test spectra.
            one_body_squared_spectrum = eigenspectrum(one_body_squared)
            two_body_spectrum = eigenspectrum(two_body_operator)
            difference = two_body_spectrum - one_body_squared_spectrum
            self.assertAlmostEqual(0., numpy.amax(numpy.absolute(difference)))

    def test_one_body_squared_nonhermitian_raises_error(self):
        one_body_matrix = numpy.array([[0, 1], [0, 0]])
        with self.assertRaises(ValueError):
            prepare_one_body_squared_evolution(one_body_matrix,
                                               spin_basis=False)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Class and functions to store interaction operators."""
import numpy

from openfermion.ops import QubitOperator
from openfermion.ops.representations import (PolynomialTensor,
                                             get_tensors_from_integrals)

COEFFICIENT_TYPES = (int, float, complex)


class DOCIHamiltonian(PolynomialTensor):
    r"""Class for storing DOCI hamiltonians which are defined to be
    restrictions of fermionic operators to doubly occupied configurations.
    As such they are by nature hard-core boson Hamiltonians, but the
    hard-core Boson algebra is identical to the Pauli algebra, which is
    why it is convenient to represent DOCI hamiltonians as QubitOperators

    Note that the operators stored in this class take the form:

        .. math::

            constant + \sum_{p} h^{(r1)}_{p, p}/2 (1 - \sigma^Z_p) +
            \sum_{p \neq q} h^{(r1)}_{p, q}/4
                * (\sigma^X_p \sigma^X_q + \sigma^Y_p \sigma^Y_q) +
            \sum_{p \neq q} h^{(r2)}_{p, q}/4 (1 - \sigma^Z_p -
            \sigma^Z_q + \sigma^Z_p \sigma^Z_q)
            =
            constant + \sum_{p} h_{p, p} N_p +
            \sum_{p \neq q} w_{p, p} N_p N_q +
            \sum_{p \neq q} v_{p, p} P_p^\dagger P_q,

    where

        .. math::

            N_p = (1 - \sigma^Z_p)/2,
            P_p = a_{i,\beta} a_{i,\alpha},
            h_p = h^{(r1)}_{p, p} = \langle p|h|p \rangle =
                2 I^{(1)}_{p, p} + I^{(2)}_{p, p, p, p},
            w_{p, q} = h^{(r2)}_{p, q} = 2 \langle pq|v|pq \rangle -
                                         \langle pq|v|qp \rangle =
                                        2 I^{(2)}_{p, q, q, p} -
                                        I^{(2)}_{p, q, p, q},
            v_{p, q} = h^{(r1)}_{p, q} = \langle pp|v|qq \rangle =
                I^{(2)}_{p, p, q, q},

    with (:math:`I^{(1)}_{p, q}`) and (:math:`I^{(2)}_{p, q, r, s}`) are the one
    and two body electron integrals and (:math:`h`) and (:math:`v`) are the
    coefficients of the corresponding InteractionOperator

        .. math::

            constant + \sum_{p, q} h_{p, q} a^\dagger_p a_q +
            \sum_{p, q, r, s} h_{p, q, r, s} a^\dagger_p a^\dagger_q a_r a_s.


    Attributes:
        constant: The constant offset.
        hr1: The coefficients of (:math:`h^{r1}_{p, q}`).
            This is an n_qubits x n_qubits numpy array of floats.
        hr2: The coefficients of (:math:`h^{r2}_{p, q}`).
            This is an n_qubits x n_qubits numpy array of floats.
    """

    def __init__(self, constant, hc, hr1, hr2):
        r"""
        Initialize the DOCIHamiltonian class.

        Args:
            constant: A constant term in the operator given as a
                float. For instance, the nuclear repulsion energy.
            hc: the coefficients of (:math:`h^{(c)}_{p}`)
            hr1: The coefficients of (:math:`h^{(r1)}_{p, q}`).
               This is an n_qubits x n_qubits numpy array of floats.
            hr2: The coefficients of (:math:`h^{(r2)}_{p, q}`).
                This is an n_qubits x n_qubits array of floats.
        """
        super(DOCIHamiltonian, self).__init__(None)

        self._n_qubits = hc.shape[0]
        self._constant = constant
        self._hr1 = hr1
        self._hr2 = hr2
        self._hc = hc

    @property
    def qubit_operator(self):
        """Return the QubitOperator representation of this DOCI Hamiltonian"""
        return self.identity_part + self.z_part + self.xy_part

    def xx_term(self, p, q):
        """Returns the XX term on a single pair of qubits as a QubitOperator
        Arguments:
            p, q [int] -- qubit indices
        Returns:
            [QubitOperator] -- XX term on the chosen qubits.
        """
        return QubitOperator("X" + str(p) + " X" + str(q), self.hr1[p, q] / 2)

    def yy_term(self, p, q):
        """Returns the YY term on a single pair of qubits as a QubitOperator
        Arguments:
            p, q [int] -- qubit indices
        Returns:
            [QubitOperator] -- YY term on the chosen qubits.
        """
        return QubitOperator("Y" + str(p) + " Y" + str(q), self.hr1[p, q] / 2)

    def z_term(self, p):
        """Returns the Z term on a single qubit as a QubitOperator
        Arguments:
            p [int] -- qubit index
        Returns:
            [QubitOperator] -- Z term on the chosen qubit.
        """
        return QubitOperator("Z" + str(p),
                             self.hc[p] / 2 + sum(self.hr2[:, p]) / 2)

    def zz_term(self, p, q):
        """Returns the ZZ term on a single pair of qubits as a QubitOperator
        Arguments:
            p, q [int] -- qubit indices
        Returns:
            [QubitOperator] -- ZZ term on the chosen qubits.
        """
        return QubitOperator("Z" + str(p) + " Z" + str(q), self.hr2[p, q] / 2)

    @property
    def identity_part(self):
        """Returns identity term of this operator (i.e. trace-ful term)
        in QubitOperator form.
        """
        return QubitOperator(
            (), self.constant + numpy.sum(self.hc) / 2 +
            numpy.sum(self.hr2) / 4 + numpy.sum(numpy.diag(self.hr2)) / 4)

    @property
    def xx_part(self):
        """Returns the XX part of the QubitOperator representation of this
        DOCIHamiltonian
        """
        return sum([
            self.xx_term(p, q)
            for p in range(self.n_qubits)
            for q in range(p + 1, self.n_qubits)
        ])

    @property
    def yy_part(self):
        """Returns the YY part of the QubitOperator representation of this
        DOCIHamiltonian
        """
        return sum([
            self.yy_term(p, q)
            for p in range(self.n_qubits)
            for q in range(p + 1, self.n_qubits)
        ])

    @property
    def xy_part(self):
        """Returns the XX+YY part of the QubitOperator representation of this
        DOCIHamiltonian
        """
        return self.xx_part + self.yy_part

    @property
    def zz_part(self):
        """Returns the ZZ part of the QubitOperator representation of this
        DOCIHamiltonian
        """
        return sum([
            self.zz_term(p, q)
            for p in range(self.n_qubits)
            for q in range(p + 1, self.n_qubits)
        ])

    @property
    def z_part(self):
        """Return the Z and ZZ part of the QubitOperator representation of this
        DOCI Hamiltonian"""
        return self.zz_part + sum(
            [self.z_term(p) for p in range(self.n_qubits)])

    @property
    def hc(self):
        return self._hc

    @hc.setter
    def hc(self, value):
        self._hc = value

    @property
    def hr1(self):
        """The value of hr1."""
        return self._hr1

    @hr1.setter
    def hr1(self, value):
        self._hr1 = value

    @property
    def hr2(self):
        """The value of hr2."""
        return self._hr2

    @hr2.setter
    def hr2(self, value):
        self._hr2 = value

    # Override base class
    @property
    def constant(self):
        return self._constant

    @constant.setter
    def constant(self, value):
        self._constant = value

    # Override base class to generate on the fly
    @property
    def n_body_tensors(self):
        one_body_coefficients, two_body_coefficients =\
            get_tensors_from_doci(self.hc, self.hr1, self.hr2)
        n_body_tensors = {
            (): self.constant,
            (1, 0): one_body_coefficients,
            (1, 1, 0, 0): two_body_coefficients
        }
        return n_body_tensors

    @n_body_tensors.setter
    def n_body_tensors(self, value):
        raise TypeError('Raw edits of the n_body_tensors of a DOCIHamiltonian '
                        'is not allowed. Either adjust the hc/hr1/hr2 terms '
                        'or cast to another PolynomialTensor class.')

    def _get_onebody_term(self, key, index):
        if key[0] != 1 or key[1] != 0:
            raise IndexError('DOCIHamiltonian class only contains '
                             'one-body terms in the (1, 0) sector.')
        if index[0] != index[1]:
            raise IndexError('DOCIHamiltonian class only contains '
                             'diagonal one-body electron integrals.')
        return self.hc[index[0] // 2] / 2

    def _get_twobody_term(self, key, index):
        if key[0] != 1 or key[1] != 1 or key[2] != 0 or key[3] != 0:
            raise IndexError('DOCIHamiltonian class only contains '
                             'two-body terms in the (1, 1, 0, 0) sector.')
        if index[0] == index[3] and index[1] == index[2]:
            return self.hr2[index[0] // 2, index[1] // 2] / 2
        if index[0] // 2 == index[1] // 2 and index[2] // 2 == index[3] // 2:
            return self.hr1[index[0] // 2, index[2] // 2] / 2
        raise IndexError('DOCIHamiltonian class only contains '
                         'two-electron integrals corresponding '
                         'to a double excitation.')

    # Override base class
    def __getitem__(self, args):
        """Look up matrix element.

        Args:
            args: Tuples indicating which coefficient to get. For instance,
                `my_tensor[(6, 1), (8, 1), (2, 0)]`
                returns
                `my_tensor.n_body_tensors[1, 1, 0][6, 8, 2]`
        """
        if len(args) == 0:
            return self.constant
        index = tuple([operator[0] for operator in args])
        key = tuple([operator[1] for operator in args])
        if len(index) == 2:
            return self._get_onebody_term(key, index)
        if len(index) == 4:
            return self._get_twobody_term(key, index)
        raise IndexError('DOCIHamiltonian class only contains '
                         'one and two-electron and constant terms.')

    # Override root class
    def __setitem__(self, args, value):
        # This is not well-defined for a DOCIHamiltonian as we want to keep
        # certain tensor elements within the class the same. Better to
        # make the user update the hr1/hr2/hc terms --- if they really
        # want to play around with the n_body_tensors here they should
        # be casting this to a raw PolynomialTensor.
        raise TypeError('Raw edits of the n_body_tensors of a DOCIHamiltonian '
                        'is not allowed. Either adjust the hc/hr1/hr2 terms '
                        'or cast to another PolynomialTensor class.')

    # Override root class
    def __iadd__(self, addend):
        if isinstance(addend, COEFFICIENT_TYPES):
            self.constant += addend
            return self
        if not issubclass(type(addend), DOCIHamiltonian):
            raise TypeError('Invalid type.')
        if self.n_qubits != addend.n_qubits:
            raise TypeError('Invalid tensor shape.')
        self.hc += addend.hc
        self.hr1 += addend.hr1
        self.hr2 += addend.hr2
        self.constant += addend.constant
        return self

    def __isub__(self, subtrahend):
        if isinstance(subtrahend, COEFFICIENT_TYPES):
            self.constant -= subtrahend
            return self
        if not issubclass(type(subtrahend), DOCIHamiltonian):
            raise TypeError('Invalid type.')
        if self.n_qubits != subtrahend.n_qubits:
            raise TypeError('Invalid tensor shape.')
        self.hc -= subtrahend.hc
        self.hr1 -= subtrahend.hr1
        self.hr2 -= subtrahend.hr2
        self.constant -= subtrahend.constant
        return self

    def __imul__(self, multiplier):
        if not isinstance(multiplier, COEFFICIENT_TYPES):
            raise TypeError('Invalid type.')
        self.hc *= multiplier
        self.hr1 *= multiplier
        self.hr2 *= multiplier
        self.constant *= multiplier
        return self

    def __itruediv__(self, dividend):
        if not isinstance(dividend, COEFFICIENT_TYPES):
            raise TypeError('Invalid type.')
        self.hc /= dividend
        self.hr1 /= dividend
        self.hr2 /= dividend
        self.constant /= dividend
        return self

    @classmethod
    def from_integrals(cls, constant, one_body_integrals, two_body_integrals):
        hc, hr1, hr2 = get_doci_from_integrals(one_body_integrals,
                                               two_body_integrals)
        return cls(constant, hc, hr1, hr2)

    @classmethod
    def zero(cls, n_qubits):
        return cls(0, numpy.zeros((n_qubits,), dtype=numpy.complex128),
                   numpy.zeros((n_qubits,) * 2, dtype=numpy.complex128),
                   numpy.zeros((n_qubits,) * 2, dtype=numpy.complex128))


def get_tensors_from_doci(hc, hr1, hr2):
    '''Makes the one and two-body tensors from the DOCI Hamiltonian matrices

    Args:
        hc [numpy array]: The single-particle DOCI terms in matrix form
        hr1 [numpy array]: The off-diagonal DOCI Hamiltonian terms in matrix
            form
        hr2 [numpy array]: The diagonal DOCI Hamiltonian terms in matrix form

    Returns:
        one_body_coefficients [numpy array]: The corresponding one-body
            tensor for the electronic structure Hamiltonian
        two_body_coefficients [numpy array]: The corresponding two body
            tensor for the electronic structure Hamiltonian
    '''
    one_body_integrals, two_body_integrals =\
        get_projected_integrals_from_doci(hc, hr1, hr2)
    one_body_coefficients, two_body_coefficients = get_tensors_from_integrals(
        one_body_integrals, two_body_integrals)
    return one_body_coefficients, two_body_coefficients


def get_projected_integrals_from_doci(hc, hr1, hr2):
    '''Generates a set of atomic integrals corresponding to this Hamiltonian

    Makes the one and two-body integrals from the DOCI projectionof the hr1,
    hr2, and hc matrices. This is technically not well-defined, as hr2 is
    not generated in a one-to-one fashion. In particular, here we assume
    hr2 corresponds entirely to a pqqp term, while this could be equally well
    added to the pqpq term. This implies that calling

    get_doci_from_integrals(*get_projected_integrals_from_doci(hc, hr1, hr2))

    should return the same hc, hr1, and hr2, but there is no such guarantee
    for

    get_projected_integrals_from_doci(*get_doci_from_integrals(
        one_body_integrals, two_body_integrals))

    Args:
        hc [numpy array]: The single-particle DOCI terms in matrix form
        hr1 [numpy array]: The off-diagonal DOCI Hamiltonian terms in matrix
            form
        hr2 [numpy array]: The diagonal DOCI Hamiltonian terms in matrix form

    Returns:
        projected_onebody_integrals [numpy array]: The corresponding one-body
            integrals for the electronic structure Hamiltonian
        projected_twobody_integrals [numpy array]: The corresponding two body
            integrals for the electronic structure Hamiltonian
    '''
    n_qubits = hr1.shape[0]
    projected_onebody_integrals = numpy.zeros((n_qubits, n_qubits))
    projected_twobody_integrals = numpy.zeros(
        (n_qubits, n_qubits, n_qubits, n_qubits))
    for p in range(n_qubits):
        projected_onebody_integrals[p, p] = hc[p] / 2
        projected_twobody_integrals[p, p, p, p] = hr2[p, p]
        for q in range(n_qubits):
            if p == q:
                continue
            projected_twobody_integrals[p, q, q, p] = hr2[p, q] / 2
            projected_twobody_integrals[p, p, q, q] = hr1[p, q]

    return projected_onebody_integrals, projected_twobody_integrals


def get_doci_from_integrals(one_body_integrals, two_body_integrals):
    r"""Construct a DOCI Hamiltonian from electron integrals

    Args:
        one_body_integrals [numpy array]: one-electron integrals
        two_body_integrals [numpy array]: two-electron integrals

    Returns:
        hc [numpy array]: The single-particle DOCI terms in matrix form
        hr1 [numpy array]: The off-diagonal DOCI Hamiltonian terms in matrix
            form
        hr2 [numpy array]: The diagonal DOCI Hamiltonian terms in matrix form
    """

    n_qubits = one_body_integrals.shape[0]
    hc = numpy.zeros(n_qubits)
    hr1 = numpy.zeros((n_qubits, n_qubits))
    hr2 = numpy.zeros((n_qubits, n_qubits))

    for p in range(n_qubits):
        hc[p] = 2 * one_body_integrals[p, p]
        for q in range(n_qubits):
            hr2[p, q] = (2 * two_body_integrals[p, q, q, p] -
                         two_body_integrals[p, q, p, q])
            if p == q:
                continue
            hr1[p, q] = two_body_integrals[p, p, q, q]

    return hc, hr1, hr2



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy
import scipy
import pytest
import cirq

import openfermion


def test_double_excitation_init_with_multiple_args_fails():
    with pytest.raises(ValueError):
        _ = openfermion.DoubleExcitationGate(exponent=1.0,
                                             duration=numpy.pi / 2)


def test_double_excitation_eq():
    eq = cirq.testing.EqualsTester()

    eq.add_equality_group(
        openfermion.DoubleExcitationGate(exponent=1.5),
        openfermion.DoubleExcitationGate(exponent=-0.5),
        openfermion.DoubleExcitationGate(rads=-0.5 * numpy.pi),
        openfermion.DoubleExcitationGate(degs=-90),
        openfermion.DoubleExcitationGate(duration=-0.5 * numpy.pi / 2))

    eq.add_equality_group(
        openfermion.DoubleExcitationGate(exponent=0.5),
        openfermion.DoubleExcitationGate(exponent=-1.5),
        openfermion.DoubleExcitationGate(rads=0.5 * numpy.pi),
        openfermion.DoubleExcitationGate(degs=90),
        openfermion.DoubleExcitationGate(duration=-1.5 * numpy.pi / 2))

    eq.make_equality_group(lambda: openfermion.DoubleExcitationGate(exponent=0.0
                                                                   ))
    eq.make_equality_group(lambda: openfermion.DoubleExcitationGate(exponent=
                                                                    0.75))


def test_double_excitation_consistency():
    openfermion.testing.assert_implements_consistent_protocols(
        openfermion.DoubleExcitation)


double_excitation_simulator_test_cases = [
    (openfermion.DoubleExcitation, 1.0,
     numpy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]) / 4.,
     numpy.array([1, 1, 1, -1, 1, 1, 1, 1, 1, 1, 1, 1, -1, 1, 1, 1]) / 4.,
     5e-6),
    (openfermion.DoubleExcitation, -1.0,
     numpy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]) / 4.,
     numpy.array([1, 1, 1, -1, 1, 1, 1, 1, 1, 1, 1, 1, -1, 1, 1, 1]) / 4.,
     5e-6),
    (openfermion.DoubleExcitation, 0.5,
     numpy.array([1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0]) /
     numpy.sqrt(8),
     numpy.array([1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1j, 0, 0, 0]) /
     numpy.sqrt(8), 5e-6),
    (openfermion.DoubleExcitation, -0.5,
     numpy.array([1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]) / 4.,
     numpy.array([1, -1, -1, -1j, -1, -1, 1, 1, 1, 1, 1, 1, 1j, 1, 1, 1]) / 4.,
     5e-6),
    (openfermion.DoubleExcitation, -1. / 7,
     numpy.array(
         [1, 1j, -1j, -1, 1, 1j, -1j, -1, 1, 1j, -1j, -1, 1, 1j, -1j, -1]) / 4.,
     numpy.array([
         1, 1j, -1j, -numpy.cos(numpy.pi / 7) - 1j * numpy.sin(numpy.pi / 7), 1,
         1j, -1j, -1, 1, 1j, -1j, -1,
         numpy.cos(numpy.pi / 7) + 1j * numpy.sin(numpy.pi / 7), 1j, -1j, -1
     ]) / 4., 5e-6),
    (openfermion.DoubleExcitation, 7. / 3,
     numpy.array([
         0, 0, 0, 2, (1 + 1j) / numpy.sqrt(2), (1 - 1j) / numpy.sqrt(2),
         -(1 + 1j) / numpy.sqrt(2), -1, 1, 1j, -1j, -1, 1, 1j, -1j, -1
     ]) / 4.,
     numpy.array([
         0, 0, 0, 1 + 1j * numpy.sqrt(3) / 2, (1 + 1j) / numpy.sqrt(2),
         (1 - 1j) / numpy.sqrt(2), -(1 + 1j) / numpy.sqrt(2), -1, 1, 1j, -1j,
         -1, 0.5 + 1j * numpy.sqrt(3), 1j, -1j, -1
     ]) / 4., 5e-6),
    (openfermion.DoubleExcitation, 0,
     numpy.array([1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]) / 4.,
     numpy.array([1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]) / 4.,
     5e-6),
    (openfermion.DoubleExcitation, 0.25,
     numpy.array([1, 0, 0, -2, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 1]) /
     numpy.sqrt(15),
     numpy.array([
         1, 0, 0, +3j / numpy.sqrt(2) - numpy.sqrt(2), 0, 0, 0, 0, 0, 0, 0, 0,
         3 / numpy.sqrt(2) - 1j * numpy.sqrt(2), 0, 0, 1
     ]) / numpy.sqrt(15), 5e-6)
]


@pytest.mark.parametrize('gate, exponent, initial_state, correct_state, atol',
                         double_excitation_simulator_test_cases)
def test_four_qubit_rotation_gates_on_simulator(gate, exponent, initial_state,
                                                correct_state, atol):

    a, b, c, d = cirq.LineQubit.range(4)
    circuit = cirq.Circuit(gate(a, b, c, d)**exponent)
    result = circuit.final_state_vector(initial_state=initial_state)
    cirq.testing.assert_allclose_up_to_global_phase(result,
                                                    correct_state,
                                                    atol=atol)


def test_double_excitation_gate_text_diagrams():
    a = cirq.NamedQubit('a')
    b = cirq.NamedQubit('b')
    c = cirq.NamedQubit('c')
    d = cirq.NamedQubit('d')

    circuit = cirq.Circuit(openfermion.DoubleExcitation(a, b, c, d))
    cirq.testing.assert_has_diagram(
        circuit, """
a: ───⇅───
      │
b: ───⇅───
      │
c: ───⇵───
      │
d: ───⇵───
""")

    circuit = cirq.Circuit(openfermion.DoubleExcitation(a, b, c, d)**-0.5)
    cirq.testing.assert_has_diagram(
        circuit, """
a: ───⇅────────
      │
b: ───⇅────────
      │
c: ───⇵────────
      │
d: ───⇵^-0.5───
""")

    circuit = cirq.Circuit(openfermion.DoubleExcitation(a, c, b, d)**0.2)
    cirq.testing.assert_has_diagram(
        circuit, """
a: ───⇅───────
      │
b: ───⇵───────
      │
c: ───⇅───────
      │
d: ───⇵^0.2───
""")

    circuit = cirq.Circuit(openfermion.DoubleExcitation(d, b, a, c)**0.7)
    cirq.testing.assert_has_diagram(
        circuit, """
a: ───⇵───────
      │
b: ───⇅───────
      │
c: ───⇵───────
      │
d: ───⇅^0.7───
""")

    circuit = cirq.Circuit(openfermion.DoubleExcitation(d, b, a, c)**2.3)
    cirq.testing.assert_has_diagram(
        circuit, """
a: ───⇵───────
      │
b: ───⇅───────
      │
c: ───⇵───────
      │
d: ───⇅^0.3───
""")


def test_double_excitation_gate_text_diagrams_no_unicode():
    a = cirq.NamedQubit('a')
    b = cirq.NamedQubit('b')
    c = cirq.NamedQubit('c')
    d = cirq.NamedQubit('d')

    circuit = cirq.Circuit(openfermion.DoubleExcitation(a, b, c, d))
    # pylint: disable=anomalous-backslash-in-string
    cirq.testing.assert_has_diagram(circuit,
                                    r"""
a: ---/\ \/---
      |
b: ---/\ \/---
      |
c: ---\/ /\---
      |
d: ---\/ /\---
""",
                                    use_unicode_characters=False)

    circuit = cirq.Circuit(openfermion.DoubleExcitation(a, b, c, d)**-0.5)
    cirq.testing.assert_has_diagram(circuit,
                                    r"""
a: ---/\ \/--------
      |
b: ---/\ \/--------
      |
c: ---\/ /\--------
      |
d: ---\/ /\^-0.5---
""",
                                    use_unicode_characters=False)

    circuit = cirq.Circuit(openfermion.DoubleExcitation(a, c, b, d)**0.2)
    cirq.testing.assert_has_diagram(circuit,
                                    r"""
a: ---/\ \/-------
      |
b: ---\/ /\-------
      |
c: ---/\ \/-------
      |
d: ---\/ /\^0.2---
""",
                                    use_unicode_characters=False)

    circuit = cirq.Circuit(openfermion.DoubleExcitation(d, b, a, c)**0.7)
    cirq.testing.assert_has_diagram(circuit,
                                    r"""
a: ---\/ /\-------
      |
b: ---/\ \/-------
      |
c: ---\/ /\-------
      |
d: ---/\ \/^0.7---
""",
                                    use_unicode_characters=False)

    circuit = cirq.Circuit(openfermion.DoubleExcitation(d, b, a, c)**2.3)
    cirq.testing.assert_has_diagram(circuit,
                                    r"""
a: ---\/ /\-------
      |
b: ---/\ \/-------
      |
c: ---\/ /\-------
      |
d: ---/\ \/^0.3---
""",
                                    use_unicode_characters=False)


@pytest.mark.parametrize('exponent', [1.0, 0.5, 0.25, 0.1, 0.0, -0.5])
def test_double_excitation_matches_fermionic_evolution(exponent):
    gate = openfermion.DoubleExcitation**exponent

    op = openfermion.FermionOperator('3^ 2^ 1 0')
    op += openfermion.hermitian_conjugated(op)
    matrix_op = openfermion.get_sparse_operator(op)

    time_evol_op = scipy.linalg.expm(-1j * matrix_op * exponent * numpy.pi)
    time_evol_op = time_evol_op.todense()

    cirq.testing.assert_allclose_up_to_global_phase(cirq.unitary(gate),
                                                    time_evol_op,
                                                    atol=1e-7)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Classes to assist in processing of VPE (ArXiv:2010.02538) signal data"""

import abc
from typing import Sequence, Optional
import numpy

import cirq

from openfermion.linalg import fit_known_frequencies
from openfermion.circuits import standard_vpe_rotation_set


class _VPEEstimator(metaclass=abc.ABCMeta):
    """Generic class for any VPE estimator"""

    @abc.abstractmethod
    def get_simulation_points(self) -> numpy.ndarray:
        """Generates time points for estimation

        VPE requires estimating the phase function g(t) at multiple points t,
        and some care in choosing these points is needed to prevent aliasing.
        This should be taken care of in the estimator.

        Returns:
            times: a set of times t that g(t) should be estimated at.
        """

    @abc.abstractmethod
    def get_expectation_value(self,
                              phase_function: numpy.ndarray) -> numpy.ndarray:
        """Estimates expectation values from an input phase function

        Given a phase function g(t), estimates the expectation value <H> of the
        operator used to generate it on the initial state rho

        $g(t) = Trace[e^{-iHt} (|psi_r><psi_r| + rho) e^{iHt} |0><1|])$

        Arguments:
            phase_function [numpy.ndarray] -- The phase function g(t)

        Returns:
            expectation value [numpy.ndarray] -- <H>.
        """


class PhaseFitEstimator(_VPEEstimator):
    """A VPE estimator that works by fitting a set of known frequencies.

    A Hamiltonian being fast-forwardable is equivalent to its spectral
    decomposition being known. This means that the only information to
    be obtained from QPE is the amplitudes. This estimator proceeds
    by a simple least-squares fit to obtain the amplitudes, and then
    outputs the expectation values.
    """

    def __init__(self, evals: numpy.ndarray, ref_eval: float = 0):
        """
        Arguments:
            evals [numpy.ndarray] -- The (known) eigenvalues of the target
                operator
            ref_eval [numpy.ndarray] -- The eigenvalue of the reference state.
                When using a control qubit for QPE, this should be set to 0.
        """
        self.evals = evals
        self.ref_eval = ref_eval

    def get_simulation_points(self, safe: bool = True) -> numpy.ndarray:
        """Generates time points for estimation

        VPE requires estimating the phase function g(t) at multiple points t,
        and some care in choosing these points is needed to prevent aliasing.
        This should be taken care of in the estimator.

        In this case, we fit len(self.energies) complex amplitudes to a complex
        valued signal, we need precisely this number of points in the signal.

        However, it appears numerically that approximately twice as many points
        are needed to prevent aliasing, so we double this number here.

        Then, to prevent aliasing, we need to make sure that the time step
        dt < 2*pi / (E_max-E_min). Here, we choose dt = pi / (E_max-E_min).
        (Importantly, for Pauli operators this reproduces the H test.)

        Args:
            safe [bool, default True] -- numerical testing shows that taking
                approximately twice as many points is better for the stability
                of the estimator; this

        Returns:
            times: a set of times t that g(t) should be estimated at.
        """
        if safe:
            numsteps = len(self.evals) * 2
            step_size = numpy.pi / (max(self.evals) - min(self.evals))
        else:
            numsteps = len(self.evals)
            step_size = numpy.pi / (max(self.evals) - min(self.evals))
        maxtime = step_size * (numsteps - 1)
        times = numpy.linspace(0, maxtime, numsteps)
        return times

    def get_amplitudes(self, phase_function: numpy.ndarray) -> numpy.ndarray:
        """Fits the amplitudes in the phase function to the input signal data.

        Arguments:
            phase_function [numpy.ndarray] -- Phase function input

        Returns:
            amplitudes [numpy.ndarray] -- Fitted estimates of the amplitudes
                of the given frequencies (in the same order as in self.energies)
        """
        times = self.get_simulation_points()
        phase_function_shifted = numpy.array(phase_function) *\
            numpy.exp(1j * times * self.ref_eval)
        amplitudes = fit_known_frequencies(phase_function_shifted, times,
                                           self.evals)
        return amplitudes

    def get_expectation_value(self,
                              phase_function: numpy.ndarray) -> numpy.ndarray:
        """Estates expectation values via amplitude fitting of known frequencies

        Arguments:
            phase_function [numpy.ndarray] -- The phase function obtained in
                experiment

        Returns:
            expectation_value [float] -- the estimated expectation value
        """
        amplitudes = self.get_amplitudes(phase_function)
        expectation_value = numpy.dot(numpy.abs(amplitudes),
                                      self.evals) / numpy.sum(
                                          numpy.abs(amplitudes))
        return expectation_value


def get_phase_function(results: Sequence[cirq.TrialResult],
                       qubits: Sequence[cirq.Qid],
                       target_qid: int,
                       rotation_set: Optional[Sequence] = None):
    """Generates an estimate of the phase function g(t) from circuit output

    The output from a VPE circuit is a set of measurements; from the frequency
    that these measurements occur, we can estimate the phase function.

    Arguments:
        measurements [Sequence[cirq.TrialResult]] -- A list of TrialResults
            from the different circuits to be run at each point. We assume that
            these circuits are correlated to the order of rotation_set, and the
            only difference should be the initial and final rotation (following)
            that data in rotation_set. We also assume that the final measurement
            is tagged with a label of 'msmt' (and that this is a measurement of
            all qubits, with the target qubit in the bit position indicated by
            target_qid)
        qubits [Sequence[cirq.Qid]] -- The list of qubits in the order that was
            passed to the final measurement call.

            Note: we flip from small endian to big endian notation within this
            function, no need to do this externally.

        target_qid [Int] -- The index of the target qubit in qubits.

            Note: we flip from small endian to big endian notation within this
            function, no need to do this externally.

        rotation_set [Sequence or None] -- The set of rotations performed to
            generate the input data in measurements. These in turn need to be
            summed together weighted by the first entry in the set (we do not
            use the other entries in the set here).

    Returns:
        phase_function [complex] -- An estimate of g(t).
    """
    hs_index = 2**(len(qubits) - target_qid - 1)
    if rotation_set is None:
        rotation_set = standard_vpe_rotation_set
    phase_function = 0
    if len(results) != len(rotation_set):
        raise ValueError("I have an incorrect number of TrialResults "
                         "in results. Correct length should be: {}".format(
                             len(rotation_set)))
    for result, rdata in zip(results, rotation_set):
        total_shots = result.data['msmt'].count()
        msmt_counts = result.data['msmt'].value_counts()
        if 0 in msmt_counts:
            vprob0 = msmt_counts[0] / total_shots
        else:
            vprob0 = 0
        if hs_index in msmt_counts:
            vprob1 = msmt_counts[hs_index] / total_shots
        else:
            vprob1 = 0
        phase_function += rdata[0] * (vprob0 - vprob1)
    return phase_function



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""A Trotter algorithm using the "fermionic simulation gate"."""

from typing import cast, Optional, Sequence, Tuple

import cirq

from openfermion.circuits.gates import (Ryxxy, Rxxyy, CRyxxy, CRxxyy, rot111,
                                        rot11)
from openfermion.ops import DiagonalCoulombHamiltonian
from openfermion.circuits.primitives import swap_network
from openfermion.circuits.trotter.trotter_algorithm import (Hamiltonian,
                                                            TrotterStep,
                                                            TrotterAlgorithm)


class LinearSwapNetworkTrotterAlgorithm(TrotterAlgorithm):
    """A Trotter algorithm using the "fermionic simulation gate".

    This algorithm simulates a DiagonalCoulombHamiltonian. It uses layers of
    fermionic swap networks to simultaneously simulate the one- and two-body
    interactions.

    This algorithm is described in arXiv:1711.04789.
    """

    supported_types = {DiagonalCoulombHamiltonian}

    def symmetric(self, hamiltonian: Hamiltonian) -> Optional[TrotterStep]:
        return SymmetricLinearSwapNetworkTrotterStep(hamiltonian)

    def asymmetric(self, hamiltonian: Hamiltonian) -> Optional[TrotterStep]:
        return AsymmetricLinearSwapNetworkTrotterStep(hamiltonian)

    def controlled_symmetric(self,
                             hamiltonian: Hamiltonian) -> Optional[TrotterStep]:
        return ControlledSymmetricLinearSwapNetworkTrotterStep(hamiltonian)

    def controlled_asymmetric(self, hamiltonian: Hamiltonian
                             ) -> Optional[TrotterStep]:
        return ControlledAsymmetricLinearSwapNetworkTrotterStep(hamiltonian)


LINEAR_SWAP_NETWORK = LinearSwapNetworkTrotterAlgorithm()


class SymmetricLinearSwapNetworkTrotterStep(TrotterStep):

    def trotter_step(self,
                     qubits: Sequence[cirq.Qid],
                     time: float,
                     control_qubit: Optional[cirq.Qid] = None) -> cirq.OP_TREE:
        n_qubits = len(qubits)

        # Apply one- and two-body interactions for half of the full time
        def one_and_two_body_interaction(p, q, a, b) -> cirq.OP_TREE:
            yield Rxxyy(0.5 * self.hamiltonian.one_body[p, q].real * time).on(
                a, b)
            yield Ryxxy(0.5 * self.hamiltonian.one_body[p, q].imag * time).on(
                a, b)
            yield rot11(rads=-self.hamiltonian.two_body[p, q] * time).on(a, b)

        yield swap_network(qubits, one_and_two_body_interaction, fermionic=True)
        qubits = qubits[::-1]

        # Apply one-body potential for the full time
        yield (cirq.rz(rads=-self.hamiltonian.one_body[i, i].real * time).on(
            qubits[i]) for i in range(n_qubits))

        # Apply one- and two-body interactions for half of the full time
        # This time, reorder the operations so that the entire Trotter step is
        # symmetric
        def one_and_two_body_interaction_reverse_order(p, q, a,
                                                       b) -> cirq.OP_TREE:
            yield rot11(rads=-self.hamiltonian.two_body[p, q] * time).on(a, b)
            yield Ryxxy(0.5 * self.hamiltonian.one_body[p, q].imag * time).on(
                a, b)
            yield Rxxyy(0.5 * self.hamiltonian.one_body[p, q].real * time).on(
                a, b)

        yield swap_network(qubits,
                           one_and_two_body_interaction_reverse_order,
                           fermionic=True,
                           offset=True)


class ControlledSymmetricLinearSwapNetworkTrotterStep(TrotterStep):

    def trotter_step(self,
                     qubits: Sequence[cirq.Qid],
                     time: float,
                     control_qubit: Optional[cirq.Qid] = None) -> cirq.OP_TREE:
        n_qubits = len(qubits)

        if not isinstance(control_qubit, cirq.Qid):
            raise TypeError('Control qudit must be specified.')

        # Apply one- and two-body interactions for half of the full time
        def one_and_two_body_interaction(p, q, a, b) -> cirq.OP_TREE:
            yield CRxxyy(0.5 * self.hamiltonian.one_body[p, q].real * time).on(
                cast(cirq.Qid, control_qubit), a, b)
            yield CRyxxy(0.5 * self.hamiltonian.one_body[p, q].imag * time).on(
                cast(cirq.Qid, control_qubit), a, b)
            yield rot111(-self.hamiltonian.two_body[p, q] * time).on(
                cast(cirq.Qid, control_qubit), a, b)

        yield swap_network(qubits, one_and_two_body_interaction, fermionic=True)
        qubits = qubits[::-1]

        # Apply one-body potential for the full time
        yield (rot11(rads=-self.hamiltonian.one_body[i, i].real * time).on(
            control_qubit, qubits[i]) for i in range(n_qubits))

        # Apply one- and two-body interactions for half of the full time
        # This time, reorder the operations so that the entire Trotter step is
        # symmetric
        def one_and_two_body_interaction_reverse_order(p, q, a,
                                                       b) -> cirq.OP_TREE:
            yield rot111(-self.hamiltonian.two_body[p, q] * time).on(
                cast(cirq.Qid, control_qubit), a, b)
            yield CRyxxy(0.5 * self.hamiltonian.one_body[p, q].imag * time).on(
                cast(cirq.Qid, control_qubit), a, b)
            yield CRxxyy(0.5 * self.hamiltonian.one_body[p, q].real * time).on(
                cast(cirq.Qid, control_qubit), a, b)

        yield swap_network(qubits,
                           one_and_two_body_interaction_reverse_order,
                           fermionic=True,
                           offset=True)

        # Apply phase from constant term
        yield cirq.rz(rads=-self.hamiltonian.constant * time).on(control_qubit)


class AsymmetricLinearSwapNetworkTrotterStep(TrotterStep):

    def trotter_step(self,
                     qubits: Sequence[cirq.Qid],
                     time: float,
                     control_qubit: Optional[cirq.Qid] = None) -> cirq.OP_TREE:
        n_qubits = len(qubits)

        # Apply one- and two-body interactions for the full time
        def one_and_two_body_interaction(p, q, a, b) -> cirq.OP_TREE:
            yield Rxxyy(self.hamiltonian.one_body[p, q].real * time).on(a, b)
            yield Ryxxy(self.hamiltonian.one_body[p, q].imag * time).on(a, b)
            yield rot11(rads=-2 * self.hamiltonian.two_body[p, q] * time).on(
                a, b)

        yield swap_network(qubits, one_and_two_body_interaction, fermionic=True)
        qubits = qubits[::-1]

        # Apply one-body potential for the full time
        yield (cirq.rz(rads=-self.hamiltonian.one_body[i, i].real * time).on(
            qubits[i]) for i in range(n_qubits))

    def step_qubit_permutation(
            self,
            qubits: Sequence[cirq.Qid],
            control_qubit: Optional[cirq.Qid] = None
    ) -> Tuple[Sequence[cirq.Qid], Optional[cirq.Qid]]:
        # A Trotter step reverses the qubit ordering
        return qubits[::-1], None

    def finish(self,
               qubits: Sequence[cirq.Qid],
               n_steps: int,
               control_qubit: Optional[cirq.Qid] = None,
               omit_final_swaps: bool = False) -> cirq.OP_TREE:
        # If the number of Trotter steps is odd, possibly swap qubits back
        if n_steps & 1 and not omit_final_swaps:
            yield swap_network(qubits, fermionic=True)


class ControlledAsymmetricLinearSwapNetworkTrotterStep(TrotterStep):

    def trotter_step(self,
                     qubits: Sequence[cirq.Qid],
                     time: float,
                     control_qubit: Optional[cirq.Qid] = None) -> cirq.OP_TREE:

        n_qubits = len(qubits)

        if not isinstance(control_qubit, cirq.Qid):
            raise TypeError('Control qudit must be specified.')

        # Apply one- and two-body interactions for the full time
        def one_and_two_body_interaction(p, q, a, b) -> cirq.OP_TREE:
            yield CRxxyy(self.hamiltonian.one_body[p, q].real * time).on(
                cast(cirq.Qid, control_qubit), a, b)
            yield CRyxxy(self.hamiltonian.one_body[p, q].imag * time).on(
                cast(cirq.Qid, control_qubit), a, b)
            yield rot111(-2 * self.hamiltonian.two_body[p, q] * time).on(
                cast(cirq.Qid, control_qubit), a, b)

        yield swap_network(qubits, one_and_two_body_interaction, fermionic=True)
        qubits = qubits[::-1]

        # Apply one-body potential for the full time
        yield (rot11(rads=-self.hamiltonian.one_body[i, i].real * time).on(
            control_qubit, qubits[i]) for i in range(n_qubits))

        # Apply phase from constant term
        yield cirq.rz(rads=-self.hamiltonian.constant * time).on(control_qubit)

    def step_qubit_permutation(
            self,
            qubits: Sequence[cirq.Qid],
            control_qubit: Optional[cirq.Qid] = None
    ) -> Tuple[Sequence[cirq.Qid], Optional[cirq.Qid]]:
        # A Trotter step reverses the qubit ordering
        return qubits[::-1], control_qubit

    def finish(self,
               qubits: Sequence[cirq.Qid],
               n_steps: int,
               control_qubit: Optional[cirq.Qid] = None,
               omit_final_swaps: bool = False) -> cirq.OP_TREE:
        # If the number of Trotter steps is odd, possibly swap qubits back
        if n_steps & 1 and not omit_final_swaps:
            yield swap_network(qubits, fermionic=True)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""This module provides generic tools for classes in ops/"""
from builtins import map, zip
import marshal
import os

import numpy
import sympy
from scipy.sparse import spmatrix

from openfermion.config import DATA_DIRECTORY, EQ_TOLERANCE
from openfermion.ops.operators import (BosonOperator, FermionOperator,
                                       MajoranaOperator, QuadOperator,
                                       QubitOperator)
from openfermion.ops.representations import (PolynomialTensor,
                                             DiagonalCoulombHamiltonian,
                                             InteractionOperator,
                                             InteractionRDM)
from openfermion.transforms.opconversions.term_reordering\
    import normal_ordered


class OperatorUtilsError(Exception):
    pass


class OperatorSpecificationError(Exception):
    pass


def hermitian_conjugated(operator):
    """Return Hermitian conjugate of operator."""
    # Handle FermionOperator
    if isinstance(operator, FermionOperator):
        conjugate_operator = FermionOperator()
        for term, coefficient in operator.terms.items():
            conjugate_term = tuple([(tensor_factor, 1 - action)
                                    for (tensor_factor,
                                         action) in reversed(term)])
            conjugate_operator.terms[conjugate_term] = coefficient.conjugate()

    # Handle BosonOperator
    elif isinstance(operator, BosonOperator):
        conjugate_operator = BosonOperator()
        for term, coefficient in operator.terms.items():
            conjugate_term = tuple([(tensor_factor, 1 - action)
                                    for (tensor_factor,
                                         action) in reversed(term)])
            # take into account that different indices commute
            conjugate_term = tuple(
                sorted(conjugate_term, key=lambda factor: factor[0]))
            conjugate_operator.terms[conjugate_term] = coefficient.conjugate()

    # Handle QubitOperator
    elif isinstance(operator, QubitOperator):
        conjugate_operator = QubitOperator()
        for term, coefficient in operator.terms.items():
            conjugate_operator.terms[term] = coefficient.conjugate()

    # Handle QuadOperator
    elif isinstance(operator, QuadOperator):
        conjugate_operator = QuadOperator()
        for term, coefficient in operator.terms.items():
            conjugate_term = reversed(term)
            # take into account that different indices commute
            conjugate_term = tuple(
                sorted(conjugate_term, key=lambda factor: factor[0]))
            conjugate_operator.terms[conjugate_term] = coefficient.conjugate()

    # Handle InteractionOperator
    elif isinstance(operator, InteractionOperator):
        conjugate_constant = operator.constant.conjugate()
        conjugate_one_body_tensor = hermitian_conjugated(
            operator.one_body_tensor)
        conjugate_two_body_tensor = hermitian_conjugated(
            operator.two_body_tensor)
        conjugate_operator = type(operator)(conjugate_constant,
                                            conjugate_one_body_tensor,
                                            conjugate_two_body_tensor)

    # Handle sparse matrix
    elif isinstance(operator, spmatrix):
        conjugate_operator = operator.getH()

    # Handle numpy array
    elif isinstance(operator, numpy.ndarray):
        conjugate_operator = operator.T.conj()

    # Unsupported type
    else:
        raise TypeError('Taking the hermitian conjugate of a {} is not '
                        'supported.'.format(type(operator).__name__))

    return conjugate_operator


def is_hermitian(operator):
    """Test if operator is Hermitian."""
    # Handle FermionOperator, BosonOperator, and InteractionOperator
    if isinstance(operator,
                  (FermionOperator, BosonOperator, InteractionOperator)):
        return (normal_ordered(operator) == normal_ordered(
            hermitian_conjugated(operator)))

    # Handle QubitOperator and QuadOperator
    if isinstance(operator, (QubitOperator, QuadOperator)):
        return operator == hermitian_conjugated(operator)

    # Handle sparse matrix
    elif isinstance(operator, spmatrix):
        difference = operator - hermitian_conjugated(operator)
        discrepancy = 0.
        if difference.nnz:
            discrepancy = max(abs(difference.data))
        return discrepancy < EQ_TOLERANCE

    # Handle numpy array
    elif isinstance(operator, numpy.ndarray):
        difference = operator - hermitian_conjugated(operator)
        discrepancy = numpy.amax(abs(difference))
        return discrepancy < EQ_TOLERANCE

    # Unsupported type
    else:
        raise TypeError('Checking whether a {} is hermitian is not '
                        'supported.'.format(type(operator).__name__))


def count_qubits(operator):
    """Compute the minimum number of qubits on which operator acts.

    Args:
        operator: FermionOperator, QubitOperator, DiagonalCoulombHamiltonian,
            or PolynomialTensor.

    Returns:
        num_qubits (int): The minimum number of qubits on which operator acts.

    Raises:
       TypeError: Operator of invalid type.
    """
    # Handle FermionOperator.
    if isinstance(operator, FermionOperator):
        num_qubits = 0
        for term in operator.terms:
            for ladder_operator in term:
                if ladder_operator[0] + 1 > num_qubits:
                    num_qubits = ladder_operator[0] + 1
        return num_qubits

    # Handle QubitOperator.
    elif isinstance(operator, QubitOperator):
        num_qubits = 0
        for term in operator.terms:
            if term:
                if term[-1][0] + 1 > num_qubits:
                    num_qubits = term[-1][0] + 1
        return num_qubits

    # Handle MajoranaOperator.
    if isinstance(operator, MajoranaOperator):
        num_qubits = 0
        for term in operator.terms:
            for majorana_index in term:
                if numpy.ceil((majorana_index + 1) / 2) > num_qubits:
                    num_qubits = int(numpy.ceil((majorana_index + 1) / 2))
        return num_qubits

    # Handle DiagonalCoulombHamiltonian
    elif isinstance(operator, DiagonalCoulombHamiltonian):
        return operator.one_body.shape[0]

    # Handle PolynomialTensor
    elif isinstance(operator, PolynomialTensor):
        return operator.n_qubits

    # Raise for other classes.
    else:
        raise TypeError('Operator of invalid type.')


def is_identity(operator):
    """Check whether QubitOperator of FermionOperator is identity.

    Args:
        operator: QubitOperator, FermionOperator,
            BosonOperator, or QuadOperator.

    Raises:
        TypeError: Operator of invalid type.
    """
    if isinstance(
            operator,
        (QubitOperator, FermionOperator, BosonOperator, QuadOperator)):
        return list(operator.terms) == [()]
    raise TypeError('Operator of invalid type.')



def get_file_path(file_name, data_directory):
    """Compute file_path for the file that stores operator.

    Args:
        file_name: The name of the saved file.
        data_directory: Optional data directory to change from default data
                        directory specified in config file.

    Returns:
        file_path (string): File path.

    Raises:
        OperatorUtilsError: File name is not provided.
    """
    if file_name:
        if file_name[-5:] != '.data':
            file_name = file_name + ".data"
    else:
        raise OperatorUtilsError("File name is not provided.")

    if data_directory is None:
        file_path = DATA_DIRECTORY + '/' + file_name
    else:
        file_path = data_directory + '/' + file_name

    return file_path


def load_operator(file_name=None, data_directory=None, plain_text=False):
    """Load FermionOperator or QubitOperator from file.

    Args:
        file_name: The name of the saved file.
        data_directory: Optional data directory to change from default data
                        directory specified in config file.
        plain_text: Whether the input file is plain text

    Returns:
        operator: The stored FermionOperator, BosonOperator,
            QuadOperator, or QubitOperator

    Raises:
        TypeError: Operator of invalid type.
    """
    file_path = get_file_path(file_name, data_directory)

    if plain_text:
        with open(file_path, 'r') as f:
            data = f.read()
            operator_type, operator_terms = data.split(":\n")

        if operator_type == 'FermionOperator':
            operator = FermionOperator(operator_terms)
        elif operator_type == 'BosonOperator':
            operator = BosonOperator(operator_terms)
        elif operator_type == 'QubitOperator':
            operator = QubitOperator(operator_terms)
        elif operator_type == 'QuadOperator':
            operator = QuadOperator(operator_terms)
        else:
            raise TypeError('Operator of invalid type.')
    else:
        with open(file_path, 'rb') as f:
            data = marshal.load(f)
            operator_type = data[0]
            operator_terms = data[1]

        if operator_type == 'FermionOperator':
            operator = FermionOperator()
            for term in operator_terms:
                operator += FermionOperator(term, operator_terms[term])
        elif operator_type == 'BosonOperator':
            operator = BosonOperator()
            for term in operator_terms:
                operator += BosonOperator(term, operator_terms[term])
        elif operator_type == 'QubitOperator':
            operator = QubitOperator()
            for term in operator_terms:
                operator += QubitOperator(term, operator_terms[term])
        elif operator_type == 'QuadOperator':
            operator = QuadOperator()
            for term in operator_terms:
                operator += QuadOperator(term, operator_terms[term])
        else:
            raise TypeError('Operator of invalid type.')

    return operator


def save_operator(operator,
                  file_name=None,
                  data_directory=None,
                  allow_overwrite=False,
                  plain_text=False):
    """Save FermionOperator or QubitOperator to file.

    Args:
        operator: An instance of FermionOperator, BosonOperator,
            or QubitOperator.
        file_name: The name of the saved file.
        data_directory: Optional data directory to change from default data
                        directory specified in config file.
        allow_overwrite: Whether to allow files to be overwritten.
        plain_text: Whether the operator should be saved to a
                        plain-text format for manual analysis

    Raises:
        OperatorUtilsError: Not saved, file already exists.
        TypeError: Operator of invalid type.
        TypeError: Coefficients in Operator sympy expressions.
    """

    file_path = get_file_path(file_name, data_directory)

    if os.path.isfile(file_path) and not allow_overwrite:
        raise OperatorUtilsError("Not saved, file already exists.")

    if isinstance(operator, FermionOperator):
        operator_type = "FermionOperator"
    elif isinstance(operator, BosonOperator):
        operator_type = "BosonOperator"
    elif isinstance(operator, QubitOperator):
        operator_type = "QubitOperator"
    elif isinstance(operator, QuadOperator):
        operator_type = "QuadOperator"
    elif isinstance(operator, (InteractionOperator, InteractionRDM)):
        raise NotImplementedError('Not yet implemented for '
                                  'InteractionOperator or InteractionRDM.')
    else:
        raise TypeError('Operator of invalid type.')

    for term in operator.terms:
        if isinstance(operator.terms[term], sympy.Expr):
            raise TypeError('Cannot save sympy expressions.')

    if plain_text:
        with open(file_path, 'w') as f:
            f.write(operator_type + ":\n" + str(operator))
    else:
        tm = operator.terms
        with open(file_path, 'wb') as f:
            marshal.dump(
                (operator_type, dict(zip(tm.keys(), map(complex,
                                                        tm.values())))), f)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Hamiltonians that are quadratic in the fermionic ladder operators."""

import warnings

import numpy
from scipy.linalg import schur

from openfermion.ops.representations import PolynomialTensor


class QuadraticHamiltonianError(Exception):
    pass


class QuadraticHamiltonian(PolynomialTensor):
    r"""Class for storing Hamiltonians that are quadratic in the fermionic
    ladder operators. The operators stored in this class take the form

    .. math::

        \sum_{p, q} (M_{pq} - \mu \delta_{pq}) a^\dagger_p a_q
        + \frac12 \sum_{p, q}
            (\Delta_{pq} a^\dagger_p a^\dagger_q + \text{h.c.})
        + \text{constant}

    where

        - :math:`M` is a Hermitian `n_qubits` x `n_qubits` matrix.
        - :math:`\Delta` is an antisymmetric `n_qubits` x `n_qubits` matrix.
        - :math:`\mu` is a real number representing the chemical potential.
        - :math:`\delta_{pq}` is the Kronecker delta symbol.

    We separate the chemical potential :math:`\mu` from :math:`M` so that
    we can use it to adjust the expectation value of the total number of
    particles.

    Attributes:
        chemical_potential(float): The chemical potential :math:`\mu`.
    """
    def __init__(self,
                 hermitian_part,
                 antisymmetric_part=None,
                 constant=0.0,
                 chemical_potential=0.0):
        r"""
        Initialize the QuadraticHamiltonian class.

        Args:
            hermitian_part(ndarray): The matrix :math:`M`, which represents the
                coefficients of the particle-number-conserving terms.
                This is an `n_qubits` x `n_qubits` numpy array of complex
                numbers.
            antisymmetric_part(ndarray): The matrix :math:`\Delta`,
                which represents the coefficients of the
                non-particle-number-conserving terms.
                This is an `n_qubits` x `n_qubits` numpy array of complex
                numbers.
            constant(float, optional): A constant term in the operator.
            chemical_potential(float, optional): The chemical potential
                :math:`\mu`.
        """
        n_qubits = hermitian_part.shape[0]

        # Initialize combined Hermitian part
        if not chemical_potential:
            combined_hermitian_part = hermitian_part
        else:
            combined_hermitian_part = (hermitian_part -
                                       chemical_potential * numpy.eye(n_qubits))

        # Initialize the PolynomialTensor
        if antisymmetric_part is None:
            super(QuadraticHamiltonian, self).__init__({
                ():
                constant,
                (1, 0):
                combined_hermitian_part
            })
        else:
            super(QuadraticHamiltonian, self).__init__({
                ():
                constant,
                (1, 0):
                combined_hermitian_part,
                (1, 1):
                0.5 * antisymmetric_part,
                (0, 0):
                -0.5 * antisymmetric_part.conj()
            })

        # Add remaining attributes
        self.chemical_potential = chemical_potential

    @property
    def combined_hermitian_part(self):
        """The Hermitian part including the chemical potential."""
        return self.n_body_tensors[1, 0]

    @property
    def antisymmetric_part(self):
        """The antisymmetric part."""
        if (1, 1) in self.n_body_tensors:
            return 2. * self.n_body_tensors[1, 1]
        else:
            return numpy.zeros((self.n_qubits, self.n_qubits), complex)

    @property
    def hermitian_part(self):
        """The Hermitian part not including the chemical potential."""
        return (self.combined_hermitian_part +
                self.chemical_potential * numpy.eye(self.n_qubits))

    @property
    def conserves_particle_number(self):
        """Whether this Hamiltonian conserves particle number."""
        discrepancy = numpy.max(numpy.abs(self.antisymmetric_part))
        return numpy.isclose(discrepancy, 0.0)

    def add_chemical_potential(self, chemical_potential):
        """Increase (or decrease) the chemical potential by some value."""
        self.n_body_tensors[1, 0] -= (chemical_potential *
                                      numpy.eye(self.n_qubits))
        self.chemical_potential += chemical_potential

    def ground_energy(self):
        """Return the ground energy."""
        orbital_energies, _, constant = (
            self.diagonalizing_bogoliubov_transform())
        return numpy.sum(
            orbital_energies[numpy.where(orbital_energies < 0.0)[0]]) + constant

    def majorana_form(self):
        r"""Return the Majorana represention of the Hamiltonian.

        Any quadratic Hamiltonian can be written in the form

        .. math::

            \frac{i}{2} \sum_{j, k} A_{jk} f_j f_k + \text{constant}

        where the :math:`f_i` are normalized Majorana fermion operators:

        .. math::

            f_j = \frac{1}{\sqrt{2}} (a^\dagger_j + a_j)

            f_{j + N} = \frac{i}{\sqrt{2}} (a^\dagger_j - a_j)

        and :math:`A` is a (2 * `n_qubits`) x (2 * `n_qubits`) real
        antisymmetric matrix. This function returns the matrix
        :math:`A` and the constant.
        """
        hermitian_part = self.combined_hermitian_part
        antisymmetric_part = self.antisymmetric_part

        # Compute the Majorana matrix using block matrix manipulations
        majorana_matrix = numpy.zeros((2 * self.n_qubits, 2 * self.n_qubits))
        # Set upper left block
        majorana_matrix[:self.n_qubits, :self.n_qubits] = numpy.real(
            -0.5j * (hermitian_part - hermitian_part.conj() +
                     antisymmetric_part - antisymmetric_part.conj()))
        # Set upper right block
        majorana_matrix[:self.n_qubits, self.n_qubits:] = numpy.real(
            0.5 * (hermitian_part + hermitian_part.conj() - antisymmetric_part -
                   antisymmetric_part.conj()))
        # Set lower left block
        majorana_matrix[self.n_qubits:, :self.n_qubits] = numpy.real(
            -0.5 * (hermitian_part + hermitian_part.conj() +
                    antisymmetric_part + antisymmetric_part.conj()))
        # Set lower right block
        majorana_matrix[self.n_qubits:, self.n_qubits:] = numpy.real(
            -0.5j * (hermitian_part - hermitian_part.conj() -
                     antisymmetric_part + antisymmetric_part.conj()))

        # Compute the constant
        majorana_constant = (0.5 * numpy.real(numpy.trace(hermitian_part)) +
                             self.n_body_tensors[()])

        return majorana_matrix, majorana_constant

    def diagonalizing_bogoliubov_transform(self, spin_sector=None):
        r"""Compute the unitary that diagonalizes a quadratic Hamiltonian.

        Any quadratic Hamiltonian can be rewritten in the form

        .. math::

            \sum_{j} \varepsilon_j b^\dagger_j b_j + \text{constant},

        where the :math:`b^\dagger_j` are a new set fermionic creation
        operators that satisfy the canonical anticommutation relations.
        The new creation operators are linear combinations of the
        original ladder operators. In the most general case, creation and
        annihilation operators are mixed together:

        .. math::

           \begin{pmatrix}
                b^\dagger_1 \\
                \vdots \\
                b^\dagger_N \\
           \end{pmatrix}
           = W
           \begin{pmatrix}
                a^\dagger_1 \\
                \vdots \\
                a^\dagger_N \\
                a_1 \\
                \vdots \\
                a_N
           \end{pmatrix},

        where :math:`W` is an :math:`N \times (2N)` matrix.
        However, if the Hamiltonian conserves particle number then
        creation operators don't need to be mixed with annihilation operators
        and :math:`W` only needs to be an :math:`N \times N` matrix:

        .. math::

           \begin{pmatrix}
                b^\dagger_1 \\
                \vdots \\
                b^\dagger_N \\
           \end{pmatrix}
           = W
           \begin{pmatrix}
                a^\dagger_1 \\
                \vdots \\
                a^\dagger_N \\
           \end{pmatrix},

        This method returns the matrix :math:`W`.

        Args:
            spin_sector (optional str): An optional integer specifying
                a spin sector to restrict to: 0 for spin-up and 1 for
                spin-down. Should only be specified if the Hamiltonian
                includes a spin degree of freedom and spin-up modes
                do not interact with spin-down modes. If specified,
                the modes are assumed to be ordered so that spin-up orbitals
                come before spin-down orbitals.

        Returns:
            orbital_energies(ndarray)
                A one-dimensional array containing the :math:`\varepsilon_j`
            diagonalizing_unitary (ndarray):
                A matrix representing the transformation :math:`W` of the
                fermionic ladder operators. If the Hamiltonian conserves
                particle number then this is :math:`N \times N`; otherwise
                it is :math:`N \times 2N`. If spin sector is specified,
                then `N` here represents the number of spatial orbitals
                rather than spin orbitals.
            constant(float)
                The constant
        """
        n_modes = self.combined_hermitian_part.shape[0]
        if spin_sector is not None and n_modes % 2:
            raise ValueError(
                'Spin sector was specified but Hamiltonian contains '
                'an odd number of modes')

        if self.conserves_particle_number:
            return self._particle_conserving_bogoliubov_transform(spin_sector)
        else:
            # TODO implement this
            if spin_sector is not None:
                raise NotImplementedError(
                    'Specifying spin sector for non-particle-conserving '
                    'Hamiltonians is not yet supported.')
            return self._non_particle_conserving_bogoliubov_transform(
                spin_sector)

    def _particle_conserving_bogoliubov_transform(self, spin_sector):
        n_modes = self.combined_hermitian_part.shape[0]
        if spin_sector is not None:
            n_sites = n_modes // 2

            def index_map(i):
                return i + spin_sector * n_sites

            spin_indices = [index_map(i) for i in range(n_sites)]
            matrix = self.combined_hermitian_part[numpy.ix_(
                spin_indices, spin_indices)]
            orbital_energies, diagonalizing_unitary_T = numpy.linalg.eigh(
                matrix)
        else:
            matrix = self.combined_hermitian_part

            if _is_spin_block_diagonal(matrix):
                up_block = matrix[:n_modes // 2, :n_modes // 2]
                down_block = matrix[n_modes // 2:, n_modes // 2:]

                up_orbital_energies, up_diagonalizing_unitary_T = (
                    numpy.linalg.eigh(up_block))
                down_orbital_energies, down_diagonalizing_unitary_T = (
                    numpy.linalg.eigh(down_block))

                orbital_energies = numpy.concatenate(
                    (up_orbital_energies, down_orbital_energies))
                diagonalizing_unitary_T = numpy.zeros((n_modes, n_modes),
                                                      dtype=complex)
                diagonalizing_unitary_T[:n_modes // 2, :n_modes //
                                        2] = up_diagonalizing_unitary_T
                diagonalizing_unitary_T[n_modes // 2:, n_modes //
                                        2:] = down_diagonalizing_unitary_T
            else:
                orbital_energies, diagonalizing_unitary_T = numpy.linalg.eigh(
                    matrix)

        return orbital_energies, diagonalizing_unitary_T.T, self.constant

    def _non_particle_conserving_bogoliubov_transform(self, spin_sector):
        majorana_matrix, majorana_constant = self.majorana_form()

        # Get the orthogonal transformation that puts majorana_matrix
        # into canonical form
        canonical, orthogonal = antisymmetric_canonical_form(majorana_matrix)
        orbital_energies = canonical[range(self.n_qubits),
                                     range(self.n_qubits, 2 * self.n_qubits)]
        constant = -0.5 * numpy.sum(orbital_energies) + majorana_constant

        # Create the matrix that converts between fermionic ladder and
        # Majorana bases
        normalized_identity = (numpy.eye(self.n_qubits, dtype=complex) /
                               numpy.sqrt(2.))
        majorana_basis_change = numpy.eye(2 * self.n_qubits,
                                          dtype=complex) / numpy.sqrt(2.)
        majorana_basis_change[self.n_qubits:, self.n_qubits:] *= -1.j
        majorana_basis_change[:self.n_qubits, self.
                              n_qubits:] = normalized_identity
        majorana_basis_change[self.n_qubits:, :self.
                              n_qubits] = 1.j * normalized_identity

        # Compute the unitary and return
        diagonalizing_unitary = majorana_basis_change.T.conj().dot(
            orthogonal.dot(majorana_basis_change))

        return orbital_energies, diagonalizing_unitary[:self.n_qubits], constant

    def diagonalizing_circuit(self):
        r"""Get a circuit for a unitary that diagonalizes this Hamiltonian

        This circuit performs the transformation to a basis in which the
        Hamiltonian takes the diagonal form

        .. math::

            \sum_{j} \varepsilon_j b^\dagger_j b_j + \text{constant}.

        Returns
        -------
            circuit_description (list[tuple]):
                A list of operations describing the circuit. Each operation
                is a tuple of objects describing elementary operations that
                can be performed in parallel. Each elementary operation
                is either the string 'pht' indicating a particle-hole
                transformation on the last fermionic mode, or a tuple of
                the form :math:`(i, j, \theta, \varphi)`,
                indicating a Givens rotation
                of modes :math:`i` and :math:`j` by angles :math:`\theta`
                and :math:`\varphi`.
        """
        # Adding inline import here to prevent circular issues
        # TODO: move this out once we have a better solution
        from openfermion.linalg.givens_rotations import (
            fermionic_gaussian_decomposition, givens_decomposition_square)
        _, transformation_matrix, _ = self.diagonalizing_bogoliubov_transform()

        if self.conserves_particle_number:
            # The Hamiltonian conserves particle number, so we don't need
            # to use the most general procedure.
            decomposition, _ = givens_decomposition_square(
                transformation_matrix)
            circuit_description = list(reversed(decomposition))
        else:
            # The Hamiltonian does not conserve particle number, so we
            # need to use the most general procedure.
            # Rearrange the transformation matrix because the circuit
            # generation routine expects it to describe annihilation
            # operators rather than creation operators.
            left_block = transformation_matrix[:, :self.n_qubits]
            right_block = transformation_matrix[:, self.n_qubits:]

            # Can't use numpy.block because that requires numpy>=1.13.0
            new_transformation_matrix = numpy.empty(
                (self.n_qubits, 2 * self.n_qubits), dtype=complex)
            new_transformation_matrix[:, :self.n_qubits] = numpy.conjugate(
                right_block)
            new_transformation_matrix[:, self.n_qubits:] = numpy.conjugate(
                left_block)

            # Get the circuit description
            decomposition, left_decomposition, _, _ = (
                fermionic_gaussian_decomposition(new_transformation_matrix))

            # need to use left_diagonal too
            circuit_description = list(
                reversed(decomposition + left_decomposition))

        return circuit_description

    # DEPRECATED FUNCTIONS
    # ====================
    def orbital_energies(self, non_negative=False):
        r"""Return the orbital energies.

        Any quadratic Hamiltonian is unitarily equivalent to a Hamiltonian
        of the form

        .. math::

            \sum_{j} \varepsilon_j b^\dagger_j b_j + \text{constant}.

        We call the :math:`\varepsilon_j` the orbital energies.
        The eigenvalues of the Hamiltonian are sums of subsets of the
        orbital energies (up to the additive constant).

        Args:
            non_negative(bool): If True, always return a list of orbital
                energies that are non-negative. This option is ignored if
                the Hamiltonian does not conserve particle number, in which
                case the returned orbital energies are always non-negative.

        Returns
        -------
        orbital_energies(ndarray)
            A one-dimensional array containing the :math:`\varepsilon_j`
        constant(float)
            The constant
        """
        warnings.warn(
            'The method `orbital_energies` is deprecated. '
            'Use the method `diagonalizing_bogoliubov_transform` '
            'instead.', DeprecationWarning)

        orbital_energies, _, constant = (
            self.diagonalizing_bogoliubov_transform())

        return orbital_energies, constant


def antisymmetric_canonical_form(antisymmetric_matrix):
    """Compute the canonical form of an antisymmetric matrix.

    The input is a real, antisymmetric n x n matrix A, where n is even.
    Its canonical form is::

        A = R^T C R

    where R is a real, orthogonal matrix and C has the form::

        [  0     D ]
        [ -D     0 ]

    where D is a diagonal matrix with nonnegative entries.

    Args:
        antisymmetric_matrix(ndarray): An antisymmetric matrix with even
            dimension.

    Returns:
        canonical(ndarray): The canonical form C of antisymmetric_matrix
        orthogonal(ndarray): The orthogonal transformation R.
    """
    # Shifted here to prevent circular import issues
    # TODO: move this out when a better solution is found.
    from openfermion.linalg.givens_rotations import swap_columns, swap_rows
    m, p = antisymmetric_matrix.shape

    if m != p or p % 2 != 0:
        raise ValueError('The input matrix must be square with even '
                         'dimension.')

    # Check that input matrix is antisymmetric
    matrix_plus_transpose = antisymmetric_matrix + antisymmetric_matrix.T
    maxval = numpy.max(numpy.abs(matrix_plus_transpose))
    if not numpy.isclose(maxval, 0.0):
        raise ValueError('The input matrix must be antisymmetric.')

    # Compute Schur decomposition
    canonical, orthogonal = schur(antisymmetric_matrix, output='real')

    # The returned form is block diagonal; we need to permute rows and columns
    # to put it into the form we want
    n = p // 2
    for i in range(1, n, 2):
        swap_rows(canonical, i, n + i - 1)
        swap_columns(canonical, i, n + i - 1)
        swap_columns(orthogonal, i, n + i - 1)
        if n % 2 != 0:
            swap_rows(canonical, n - 1, n + i)
            swap_columns(canonical, n - 1, n + i)
            swap_columns(orthogonal, n - 1, n + i)

    # Now we permute so that the upper right block is non-negative
    for i in range(n):
        if canonical[i, n + i] < 0.0:
            swap_rows(canonical, i, n + i)
            swap_columns(canonical, i, n + i)
            swap_columns(orthogonal, i, n + i)

    # Now we permute so that the nonzero entries are ordered by magnitude
    # We use insertion sort
    diagonal = canonical[range(n), range(n, 2 * n)]
    for i in range(n):
        # Insert the smallest element from the unsorted part of the list into
        # index i
        arg_min = numpy.argmin(diagonal[i:]) + i
        if arg_min != i:
            # Permute the upper right block
            swap_rows(canonical, i, arg_min)
            swap_columns(canonical, n + i, n + arg_min)
            swap_columns(orthogonal, n + i, n + arg_min)
            # Permute the lower left block
            swap_rows(canonical, n + i, n + arg_min)
            swap_columns(canonical, i, arg_min)
            swap_columns(orthogonal, i, arg_min)
            # Update diagonal
            swap_rows(diagonal, i, arg_min)

    return canonical, orthogonal.T


def _is_spin_block_diagonal(matrix):
    n = matrix.shape[0]
    if n % 2:
        return False
    max_upper_right = numpy.max(numpy.abs(matrix[:n // 2, n // 2:]))
    max_lower_left = numpy.max(numpy.abs(matrix[n // 2:, :n // 2]))
    return (numpy.isclose(max_upper_right, 0.0) and
            numpy.isclose(max_lower_left, 0.0))



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests  jordan_wigner.py."""
import itertools
import os
import unittest

import numpy
import sympy

from openfermion.config import DATA_DIRECTORY
from openfermion.hamiltonians import fermi_hubbard, number_operator
from openfermion.chem import MolecularData
from openfermion.ops.operators import (FermionOperator, MajoranaOperator,
                                       QubitOperator)
from openfermion.ops.representations import InteractionOperator
from openfermion.transforms.opconversions import (get_fermion_operator,
                                                  reverse_jordan_wigner)
from openfermion.transforms.repconversions import (
    get_diagonal_coulomb_hamiltonian, get_interaction_operator)
from openfermion.utils import hermitian_conjugated
from openfermion.transforms.opconversions import normal_ordered
from openfermion.testing.testing_utils import (random_interaction_operator,
                                               random_quadratic_hamiltonian)

from openfermion.transforms.opconversions.jordan_wigner import (
    jordan_wigner, jordan_wigner_one_body, jordan_wigner_two_body,
    _jordan_wigner_interaction_op)


class JordanWignerTransformTest(unittest.TestCase):

    def setUp(self):
        self.n_qubits = 5

    def test_bad_input(self):
        with self.assertRaises(TypeError):
            jordan_wigner(3)

    def test_transm_raise3(self):
        raising = jordan_wigner(FermionOperator(((3, 1),)))
        self.assertEqual(len(raising.terms), 2)

        correct_operators_x = ((0, 'Z'), (1, 'Z'), (2, 'Z'), (3, 'X'))
        correct_operators_y = ((0, 'Z'), (1, 'Z'), (2, 'Z'), (3, 'Y'))
        qtermx = QubitOperator(correct_operators_x, 0.5)
        qtermy = QubitOperator(correct_operators_y, -0.5j)

        self.assertEqual(raising.terms[correct_operators_x], 0.5)
        self.assertEqual(raising.terms[correct_operators_y], -0.5j)
        self.assertTrue(raising == qtermx + qtermy)

    def test_transm_hopping_sympy(self):
        coeff = sympy.Symbol('x')
        hopping = FermionOperator(((3, 1), (2, 0))) * coeff
        hopping += FermionOperator(((2, 1), (3, 0))) * coeff

        jw_hopping = jordan_wigner(hopping)
        print(jw_hopping)
        correct_operators_xx = ((2, 'X'), (3, 'X'))
        correct_operators_yy = ((2, 'Y'), (3, 'Y'))
        qtermxx = QubitOperator(correct_operators_xx, 0.5 * coeff)
        qtermyy = QubitOperator(correct_operators_yy, 0.5 * coeff)

        self.assertEqual(jw_hopping.terms[correct_operators_xx], 0.5 * coeff)
        self.assertEqual(jw_hopping.terms[correct_operators_yy], 0.5 * coeff)
        self.assertTrue(jw_hopping == qtermxx + qtermyy)

    def test_transm_raise3_sympy(self):
        coeff = sympy.Symbol('x')
        raising = jordan_wigner(coeff * FermionOperator(((3, 1),)))
        self.assertEqual(len(raising.terms), 2)

        correct_operators_x = ((0, 'Z'), (1, 'Z'), (2, 'Z'), (3, 'X'))
        correct_operators_y = ((0, 'Z'), (1, 'Z'), (2, 'Z'), (3, 'Y'))
        qtermx = QubitOperator(correct_operators_x, 0.5 * coeff)
        qtermy = QubitOperator(correct_operators_y, -0.5j * coeff)

        self.assertEqual(raising.terms[correct_operators_x], 0.5 * coeff)
        self.assertEqual(raising.terms[correct_operators_y], -0.5j * coeff)
        self.assertTrue(raising == qtermx + qtermy)

    def test_transm_raise1(self):
        raising = jordan_wigner(FermionOperator(((1, 1),)))

        correct_operators_x = ((0, 'Z'), (1, 'X'))
        correct_operators_y = ((0, 'Z'), (1, 'Y'))
        qtermx = QubitOperator(correct_operators_x, 0.5)
        qtermy = QubitOperator(correct_operators_y, -0.5j)

        self.assertEqual(raising.terms[correct_operators_x], 0.5)
        self.assertEqual(raising.terms[correct_operators_y], -0.5j)
        self.assertTrue(raising == qtermx + qtermy)

    def test_transm_lower3(self):
        lowering = jordan_wigner(FermionOperator(((3, 0),)))

        correct_operators_x = ((0, 'Z'), (1, 'Z'), (2, 'Z'), (3, 'X'))
        correct_operators_y = ((0, 'Z'), (1, 'Z'), (2, 'Z'), (3, 'Y'))
        qtermx = QubitOperator(correct_operators_x, 0.5)
        qtermy = QubitOperator(correct_operators_y, 0.5j)

        self.assertEqual(lowering.terms[correct_operators_x], 0.5)
        self.assertEqual(lowering.terms[correct_operators_y], 0.5j)
        self.assertTrue(lowering == qtermx + qtermy)

    def test_transm_lower2(self):
        lowering = jordan_wigner(FermionOperator(((2, 0),)))

        correct_operators_x = ((0, 'Z'), (1, 'Z'), (2, 'X'))
        correct_operators_y = ((0, 'Z'), (1, 'Z'), (2, 'Y'))
        qtermx = QubitOperator(correct_operators_x, 0.5)
        qtermy = QubitOperator(correct_operators_y, 0.5j)

        self.assertEqual(lowering.terms[correct_operators_x], 0.5)
        self.assertEqual(lowering.terms[correct_operators_y], 0.5j)
        self.assertTrue(lowering == qtermx + qtermy)

    def test_transm_lower1(self):
        lowering = jordan_wigner(FermionOperator(((1, 0),)))

        correct_operators_x = ((0, 'Z'), (1, 'X'))
        correct_operators_y = ((0, 'Z'), (1, 'Y'))
        qtermx = QubitOperator(correct_operators_x, 0.5)
        qtermy = QubitOperator(correct_operators_y, 0.5j)

        self.assertEqual(lowering.terms[correct_operators_x], 0.5)
        self.assertEqual(lowering.terms[correct_operators_y], 0.5j)
        self.assertTrue(lowering == qtermx + qtermy)

    def test_transm_lower0(self):
        lowering = jordan_wigner(FermionOperator(((0, 0),)))

        correct_operators_x = ((0, 'X'),)
        correct_operators_y = ((0, 'Y'),)
        qtermx = QubitOperator(correct_operators_x, 0.5)
        qtermy = QubitOperator(correct_operators_y, 0.5j)

        self.assertEqual(lowering.terms[correct_operators_x], 0.5)
        self.assertEqual(lowering.terms[correct_operators_y], 0.5j)
        self.assertTrue(lowering == qtermx + qtermy)

    def test_transm_raise3lower0(self):
        # recall that creation gets -1j on Y and annihilation gets +1j on Y.
        term = jordan_wigner(FermionOperator(((3, 1), (0, 0))))
        self.assertEqual(term.terms[((0, 'X'), (1, 'Z'), (2, 'Z'), (3, 'Y'))],
                         0.25 * 1 * -1j)
        self.assertEqual(term.terms[((0, 'Y'), (1, 'Z'), (2, 'Z'), (3, 'Y'))],
                         0.25 * 1j * -1j)
        self.assertEqual(term.terms[((0, 'Y'), (1, 'Z'), (2, 'Z'), (3, 'X'))],
                         0.25 * 1j * 1)
        self.assertEqual(term.terms[((0, 'X'), (1, 'Z'), (2, 'Z'), (3, 'X'))],
                         0.25 * 1 * 1)

    def test_transm_number(self):
        n = number_operator(self.n_qubits, 3)
        n_jw = jordan_wigner(n)
        self.assertEqual(n_jw.terms[((3, 'Z'),)], -0.5)
        self.assertEqual(n_jw.terms[()], 0.5)
        self.assertEqual(len(n_jw.terms), 2)

    def test_ccr_offsite_even_ca(self):
        c2 = FermionOperator(((2, 1),))
        a4 = FermionOperator(((4, 0),))

        self.assertTrue(normal_ordered(c2 * a4) == normal_ordered(-a4 * c2))
        self.assertTrue(jordan_wigner(c2 * a4) == jordan_wigner(-a4 * c2))

    def test_ccr_offsite_odd_ca(self):
        c1 = FermionOperator(((1, 1),))
        a4 = FermionOperator(((4, 0),))
        self.assertTrue(normal_ordered(c1 * a4) == normal_ordered(-a4 * c1))

        self.assertTrue(jordan_wigner(c1 * a4) == jordan_wigner(-a4 * c1))

    def test_ccr_offsite_even_cc(self):
        c2 = FermionOperator(((2, 1),))
        c4 = FermionOperator(((4, 1),))
        self.assertTrue(normal_ordered(c2 * c4) == normal_ordered(-c4 * c2))

        self.assertTrue(jordan_wigner(c2 * c4) == jordan_wigner(-c4 * c2))

    def test_ccr_offsite_odd_cc(self):
        c1 = FermionOperator(((1, 1),))
        c4 = FermionOperator(((4, 1),))
        self.assertTrue(normal_ordered(c1 * c4) == normal_ordered(-c4 * c1))

        self.assertTrue(jordan_wigner(c1 * c4) == jordan_wigner(-c4 * c1))

    def test_ccr_offsite_even_aa(self):
        a2 = FermionOperator(((2, 0),))
        a4 = FermionOperator(((4, 0),))
        self.assertTrue(normal_ordered(a2 * a4) == normal_ordered(-a4 * a2))

        self.assertTrue(jordan_wigner(a2 * a4) == jordan_wigner(-a4 * a2))

    def test_ccr_offsite_odd_aa(self):
        a1 = FermionOperator(((1, 0),))
        a4 = FermionOperator(((4, 0),))
        self.assertTrue(normal_ordered(a1 * a4) == normal_ordered(-a4 * a1))

        self.assertTrue(jordan_wigner(a1 * a4) == jordan_wigner(-a4 * a1))

    def test_ccr_onsite(self):
        c1 = FermionOperator(((1, 1),))
        a1 = hermitian_conjugated(c1)
        self.assertTrue(
            normal_ordered(c1 * a1) == FermionOperator(()) -
            normal_ordered(a1 * c1))
        self.assertTrue(
            jordan_wigner(c1 * a1) == QubitOperator(()) -
            jordan_wigner(a1 * c1))

    def test_jordan_wigner_transm_op(self):
        n = number_operator(self.n_qubits)
        n_jw = jordan_wigner(n)
        self.assertEqual(self.n_qubits + 1, len(n_jw.terms))
        self.assertEqual(self.n_qubits / 2., n_jw.terms[()])
        for qubit in range(self.n_qubits):
            operators = ((qubit, 'Z'),)
            self.assertEqual(n_jw.terms[operators], -0.5)


class InteractionOperatorsJWTest(unittest.TestCase):

    def setUp(self):
        self.n_qubits = 5
        self.constant = 0.
        self.one_body = numpy.zeros((self.n_qubits, self.n_qubits), float)
        self.two_body = numpy.zeros(
            (self.n_qubits, self.n_qubits, self.n_qubits, self.n_qubits), float)
        self.interaction_operator = InteractionOperator(self.constant,
                                                        self.one_body,
                                                        self.two_body)

    def test_consistency(self):
        """Test consistency with JW for FermionOperators."""
        # Random interaction operator
        n_qubits = 5
        iop = random_interaction_operator(n_qubits, real=False)
        op1 = jordan_wigner(iop)
        op2 = jordan_wigner(get_fermion_operator(iop))

        self.assertEqual(op1, op2)

        # Interaction operator from molecule
        geometry = [('Li', (0., 0., 0.)), ('H', (0., 0., 1.45))]
        basis = 'sto-3g'
        multiplicity = 1

        filename = os.path.join(DATA_DIRECTORY, 'H1-Li1_sto-3g_singlet_1.45')
        molecule = MolecularData(geometry,
                                 basis,
                                 multiplicity,
                                 filename=filename)
        molecule.load()

        iop = molecule.get_molecular_hamiltonian()
        op1 = jordan_wigner(iop)
        op2 = jordan_wigner(get_fermion_operator(iop))

        self.assertEqual(op1, op2)

    def testjordan_wigner_one_body(self):
        # Make sure it agrees with jordan_wigner(FermionTerm).
        for p in range(self.n_qubits):
            for q in range(self.n_qubits):
                coefficient = numpy.random.randn()
                if p != q:
                    coefficient += 1j * numpy.random.randn()
                # Get test qubit operator.
                test_operator = jordan_wigner_one_body(p, q, coefficient)

                # Get correct qubit operator.
                fermion_term = FermionOperator(((p, 1), (q, 0)), coefficient)
                if p != q:
                    fermion_term += FermionOperator(((q, 1), (p, 0)),
                                                    coefficient.conjugate())
                correct_op = jordan_wigner(fermion_term)

                self.assertTrue(test_operator == correct_op)

    def testjordan_wigner_two_body(self):
        # Make sure it agrees with jordan_wigner(FermionTerm).
        for p, q, r, s in itertools.product(range(self.n_qubits), repeat=4):
            coefficient = numpy.random.randn()
            if set([p, q]) != set([r, s]):
                coefficient += 1j * numpy.random.randn()

            # Get test qubit operator.
            test_operator = jordan_wigner_two_body(p, q, r, s, coefficient)

            # Get correct qubit operator.
            fermion_term = FermionOperator(((p, 1), (q, 1), (r, 0), (s, 0)),
                                           coefficient)
            if set([p, q]) != set([r, s]):
                fermion_term += FermionOperator(
                    ((s, 1), (r, 1), (q, 0), (p, 0)), coefficient.conjugate())
            correct_op = jordan_wigner(fermion_term)

            self.assertTrue(test_operator == correct_op,
                            str(test_operator - correct_op))

    def test_jordan_wigner_twobody_interaction_op_allunique(self):
        test_op = FermionOperator('1^ 2^ 3 4')
        test_op += hermitian_conjugated(test_op)

        retransformed_test_op = reverse_jordan_wigner(
            jordan_wigner(get_interaction_operator(test_op)))

        self.assertTrue(
            normal_ordered(retransformed_test_op) == normal_ordered(test_op))

    def test_jordan_wigner_twobody_interaction_op_reversal_symmetric(self):
        test_op = FermionOperator('1^ 2^ 2 1')
        test_op += hermitian_conjugated(test_op)
        self.assertTrue(
            jordan_wigner(test_op) == jordan_wigner(
                get_interaction_operator(test_op)))

    def test_jordan_wigner_interaction_op_too_few_n_qubits(self):
        with self.assertRaises(ValueError):
            _jordan_wigner_interaction_op(self.interaction_operator,
                                          self.n_qubits - 2)

    def test_jordan_wigner_interaction_op_with_zero_term(self):
        test_op = FermionOperator('1^ 2^ 3 4')
        test_op += hermitian_conjugated(test_op)

        interaction_op = get_interaction_operator(test_op)
        interaction_op.constant = 0.0

        retransformed_test_op = reverse_jordan_wigner(
            jordan_wigner(interaction_op))

        self.assertEqual(normal_ordered(retransformed_test_op),
                         normal_ordered(test_op))


class GetInteractionOperatorTest(unittest.TestCase):

    def setUp(self):
        self.n_qubits = 5
        self.constant = 0.
        self.one_body = numpy.zeros((self.n_qubits, self.n_qubits), float)
        self.two_body = numpy.zeros(
            (self.n_qubits, self.n_qubits, self.n_qubits, self.n_qubits), float)

    def test_get_interaction_operator_identity(self):
        interaction_operator = InteractionOperator(-2j, self.one_body,
                                                   self.two_body)
        qubit_operator = jordan_wigner(interaction_operator)
        self.assertTrue(qubit_operator == -2j * QubitOperator(()))
        self.assertEqual(
            interaction_operator,
            get_interaction_operator(reverse_jordan_wigner(qubit_operator),
                                     self.n_qubits))

    def test_get_interaction_operator_one_body(self):
        interaction_operator = get_interaction_operator(FermionOperator('2^ 2'),
                                                        self.n_qubits)
        one_body = numpy.zeros((self.n_qubits, self.n_qubits), float)
        one_body[2, 2] = 1.
        self.assertEqual(interaction_operator,
                         InteractionOperator(0.0, one_body, self.two_body))

    def test_get_interaction_operator_one_body_twoterm(self):
        interaction_operator = get_interaction_operator(
            FermionOperator('2^ 3', -2j) + FermionOperator('3^ 2', 3j),
            self.n_qubits)
        one_body = numpy.zeros((self.n_qubits, self.n_qubits), complex)
        one_body[2, 3] = -2j
        one_body[3, 2] = 3j
        self.assertEqual(interaction_operator,
                         InteractionOperator(0.0, one_body, self.two_body))

    def test_get_interaction_operator_two_body(self):
        interaction_operator = get_interaction_operator(
            FermionOperator('2^ 2 3^ 4'), self.n_qubits)
        two_body = numpy.zeros(
            (self.n_qubits, self.n_qubits, self.n_qubits, self.n_qubits), float)
        two_body[3, 2, 4, 2] = -1.
        self.assertEqual(interaction_operator,
                         InteractionOperator(0.0, self.one_body, two_body))

    def test_get_interaction_operator_two_body_distinct(self):
        interaction_operator = get_interaction_operator(
            FermionOperator('0^ 1^ 2 3'), self.n_qubits)
        two_body = numpy.zeros(
            (self.n_qubits, self.n_qubits, self.n_qubits, self.n_qubits), float)
        two_body[1, 0, 3, 2] = 1.
        self.assertEqual(interaction_operator,
                         InteractionOperator(0.0, self.one_body, two_body))


class JordanWignerDiagonalCoulombHamiltonianTest(unittest.TestCase):

    def test_hubbard(self):
        x_dim = 4
        y_dim = 5
        tunneling = 2.
        coulomb = 3.
        chemical_potential = 7.
        magnetic_field = 11.
        periodic = False

        hubbard_model = fermi_hubbard(x_dim, y_dim, tunneling, coulomb,
                                      chemical_potential, magnetic_field,
                                      periodic)

        self.assertTrue(
            jordan_wigner(hubbard_model) == jordan_wigner(
                get_diagonal_coulomb_hamiltonian(hubbard_model)))

    def test_random_quadratic(self):
        n_qubits = 5
        quad_ham = random_quadratic_hamiltonian(n_qubits, True)
        ferm_op = get_fermion_operator(quad_ham)
        self.assertTrue(
            jordan_wigner(ferm_op) == jordan_wigner(
                get_diagonal_coulomb_hamiltonian(ferm_op)))


def test_jordan_wigner_majorana_op_consistent():
    op = (MajoranaOperator((1, 3, 4), 0.5) + MajoranaOperator(
        (3, 7, 8, 9, 10, 12), 1.8) + MajoranaOperator((0, 4)))
    assert jordan_wigner(op) == jordan_wigner(get_fermion_operator(op))



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""tests for fourier_transforms.py"""

import unittest

from openfermion.utils import Grid
from openfermion.hamiltonians import plane_wave_hamiltonian
from openfermion.transforms.opconversions import normal_ordered
from openfermion.linalg import get_sparse_operator
from openfermion.utils import is_hermitian

from openfermion.transforms.repconversions.fourier_transforms import (
    fourier_transform, inverse_fourier_transform)


class FourierTransformTest(unittest.TestCase):

    def test_fourier_transform(self):
        for length in [2, 3]:
            grid = Grid(dimensions=1, scale=1.5, length=length)
            spinless_set = [True, False]
            geometry = [('H', (0.1,)), ('H', (0.5,))]
            for spinless in spinless_set:
                h_plane_wave = plane_wave_hamiltonian(grid, geometry, spinless,
                                                      True)
                h_dual_basis = plane_wave_hamiltonian(grid, geometry, spinless,
                                                      False)
                h_plane_wave_t = fourier_transform(h_plane_wave, grid, spinless)

                self.assertEqual(normal_ordered(h_plane_wave_t),
                                 normal_ordered(h_dual_basis))

                # Verify that all 3 are Hermitian
                plane_wave_operator = get_sparse_operator(h_plane_wave)
                dual_operator = get_sparse_operator(h_dual_basis)
                plane_wave_t_operator = get_sparse_operator(h_plane_wave_t)
                self.assertTrue(is_hermitian(plane_wave_operator))
                self.assertTrue(is_hermitian(dual_operator))
                self.assertTrue(is_hermitian(plane_wave_t_operator))

    def test_inverse_fourier_transform_1d(self):
        grid = Grid(dimensions=1, scale=1.5, length=4)
        spinless_set = [True, False]
        geometry = [('H', (0,)), ('H', (0.5,))]
        for spinless in spinless_set:
            h_plane_wave = plane_wave_hamiltonian(grid, geometry, spinless,
                                                  True)
            h_dual_basis = plane_wave_hamiltonian(grid, geometry, spinless,
                                                  False)
            h_dual_basis_t = inverse_fourier_transform(h_dual_basis, grid,
                                                       spinless)
            self.assertEqual(normal_ordered(h_dual_basis_t),
                             normal_ordered(h_plane_wave))

    def test_inverse_fourier_transform_2d(self):
        grid = Grid(dimensions=2, scale=1.5, length=3)
        spinless = True
        geometry = [('H', (0, 0)), ('H', (0.5, 0.8))]
        h_plane_wave = plane_wave_hamiltonian(grid, geometry, spinless, True)
        h_dual_basis = plane_wave_hamiltonian(grid, geometry, spinless, False)
        h_dual_basis_t = inverse_fourier_transform(h_dual_basis, grid, spinless)
        self.assertEqual(normal_ordered(h_dual_basis_t),
                         normal_ordered(h_plane_wave))


#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pytest
from scipy.linalg import expm, kron
import cirq

import openfermion


def test_fswap_interchangeable():
    a, b = cirq.LineQubit.range(2)
    assert openfermion.FSWAP(a, b) == openfermion.FSWAP(b, a)


def test_fswap_num_qubits():
    assert openfermion.FSWAP.num_qubits() == 2


def test_fswap_inverse():
    assert openfermion.FSWAP**-1 == openfermion.FSWAP


def test_fswap_str():
    assert str(openfermion.FSWAP) == 'FSWAP'
    assert str(openfermion.FSWAP**0.5) == 'FSWAP**0.5'
    assert str(openfermion.FSWAP**-0.25) == 'FSWAP**-0.25'


def test_fswap_repr():
    assert repr(openfermion.FSWAP) == 'openfermion.FSWAP'
    assert repr(openfermion.FSWAP**0.5) == '(openfermion.FSWAP**0.5)'
    assert repr(openfermion.FSWAP**-0.25) == '(openfermion.FSWAP**-0.25)'

    openfermion.testing.assert_equivalent_repr(openfermion.FSWAP)
    openfermion.testing.assert_equivalent_repr(openfermion.FSWAP**0.5)
    openfermion.testing.assert_equivalent_repr(openfermion.FSWAP**-0.25)


def test_fswap_matrix():
    np.testing.assert_allclose(
        cirq.unitary(openfermion.FSWAP),
        np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, -1]]))

    np.testing.assert_allclose(
        cirq.unitary(openfermion.FSWAP**0.5),
        np.array([[1, 0, 0, 0], [0, 0.5 + 0.5j, 0.5 - 0.5j, 0],
                  [0, 0.5 - 0.5j, 0.5 + 0.5j, 0], [0, 0, 0, 1j]]))

    cirq.testing.assert_has_consistent_apply_unitary_for_various_exponents(
        val=openfermion.FSWAP, exponents=[1, 0.5])


@pytest.mark.parametrize('rads', [
    2 * np.pi, np.pi, 0.5 * np.pi, 0.25 * np.pi, 0.1 * np.pi, 0.0, -0.5 * np.pi
])
def test_compare_ryxxy_to_cirq_equivalent(rads):
    old_gate = openfermion.Ryxxy(rads=rads)
    new_gate = cirq.givens(angle_rads=rads)
    np.testing.assert_allclose(cirq.unitary(old_gate), cirq.unitary(new_gate))


@pytest.mark.parametrize('rads', [
    2 * np.pi, np.pi, 0.5 * np.pi, 0.25 * np.pi, 0.1 * np.pi, 0.0, -0.5 * np.pi
])
def test_rxxyy_unitary(rads):
    X = np.array([[0, 1], [1, 0]])
    Y = np.array([[0, -1j], [1j, 0]])
    XX = kron(X, X)
    YY = kron(Y, Y)
    np.testing.assert_allclose(cirq.unitary(openfermion.Rxxyy(rads)),
                               expm(-1j * rads * (XX + YY) / 2),
                               atol=1e-8)


@pytest.mark.parametrize('rads', [
    2 * np.pi, np.pi, 0.5 * np.pi, 0.25 * np.pi, 0.1 * np.pi, 0.0, -0.5 * np.pi
])
def test_ryxxy_unitary(rads):
    X = np.array([[0, 1], [1, 0]])
    Y = np.array([[0, -1j], [1j, 0]])
    YX = kron(Y, X)
    XY = kron(X, Y)
    np.testing.assert_allclose(cirq.unitary(openfermion.Ryxxy(rads)),
                               expm(-1j * rads * (YX - XY) / 2),
                               atol=1e-8)


@pytest.mark.parametrize('rads', [
    2 * np.pi, np.pi, 0.5 * np.pi, 0.25 * np.pi, 0.1 * np.pi, 0.0, -0.5 * np.pi
])
def test_rzz_unitary(rads):
    ZZ = np.diag([1, -1, -1, 1])
    np.testing.assert_allclose(cirq.unitary(openfermion.Rzz(rads)),
                               expm(-1j * ZZ * rads))


def test_common_gate_text_diagrams():
    a = cirq.NamedQubit('a')
    b = cirq.NamedQubit('b')

    circuit = cirq.Circuit(openfermion.FSWAP(a, b),
                           openfermion.FSWAP(a, b)**0.5)
    cirq.testing.assert_has_diagram(
        circuit, """
a: ───×ᶠ───×ᶠ───────
      │    │
b: ───×ᶠ───×ᶠ^0.5───
""")

    cirq.testing.assert_has_diagram(circuit,
                                    """
a: ---fswap---fswap-------
      |       |
b: ---fswap---fswap^0.5---
""",
                                    use_unicode_characters=False)
